import json
import logging
import re
from importlib.resources import files
from typing import Any

from deepdiff import DeepDiff
from pydantic import BaseModel, Field, ConfigDict, PrivateAttr

from ipfabric.models import Policy, Role

logger = logging.getLogger("ipfabric")

EXCLUDE_PATH = re.compile(r"\['(policy_id|role_id)'\]")


class ManagedRBAC(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    ipf: Any = Field(exclude=True)
    _current_policies: dict = PrivateAttr(default_factory=dict)
    _policies: dict = PrivateAttr(default_factory=dict)
    _current_roles: dict = PrivateAttr(default_factory=dict)
    _roles: dict = PrivateAttr(default_factory=dict)

    def model_post_init(self, __context: Any) -> None:
        path = files("ipfabric.tools.managed_rbac")
        if not (path.joinpath("policies.json").is_file() and path.joinpath("roles.json").is_file()):
            logger.critical("No managed RBAC found for this version, please contact your SA.")
            exit()

        self._policies = json.loads(path.joinpath("policies.json").read_text())
        self._current_policies = {
            k: {"description": v.description, "paths": v.api_scope_ids, "policy_id": v.policy_id}
            for k, v in self.ipf.settings.policies.policies_by_name.items()
            if k in self._policies
        }
        self._roles = json.loads(path.joinpath("roles.json").read_text())
        self._current_roles = {
            k: {"description": v.description, "policies": [p.name for p in v.policies], "role_id": v.role_id}
            for k, v in self.ipf.settings.roles.roles_by_name.items()
            if k in self._roles
        }

    def _apply_policy(self) -> dict[str, Policy]:
        policies = dict()
        if not self._current_policies:
            logger.warning("No Managed RBAC Policies found, pushing new policies.")
            policies = self.ipf.settings.policies.create_policies(
                [
                    {"name": k, "description": v["description"], "api_scope_ids": v["paths"]}
                    for k, v in self._policies.items()
                ]
            )
            return policies
        missing = {k: v for k, v in self._policies.items() if k not in self._current_policies}
        configured = {k: v for k, v in self._policies.items() if k in self._current_policies}
        if missing:
            logger.warning("Missing Managed RBAC Policies found, pushing new policies.")
            policies.update(
                self.ipf.settings.policies.create_policies(
                    [
                        {"name": k, "description": v["description"], "api_scope_ids": v["paths"]}
                        for k, v in missing.items()
                    ]
                )
            )
        if configured:
            diff = DeepDiff(configured, self._current_policies, exclude_regex_paths=EXCLUDE_PATH, ignore_order=True)
            if diff:
                logger.warning("Policy changes found, pushing updates.")
                for _ in diff.affected_root_keys.items:
                    logger.warning(f"Updating policy: {_}.")
                    policy = self.ipf.settings.policies.update_policy(
                        policy=self._current_policies[_]["policy_id"],
                        api_scope_ids=self._policies[_]["paths"],
                        name=_,
                        description=self._policies[_]["description"],
                        replace_api_scopes=True,
                    )
                    policies.update({_: policy})
                policies.update(
                    {
                        _: self.ipf.settings.policies.policies_by_name[_]
                        for _ in configured
                        if _ not in diff.affected_root_keys.items
                    }
                )
            else:
                policies.update({_: self.ipf.settings.policies.policies_by_name[_] for _ in configured})
        return policies

    def _apply_role(self, policies) -> dict[str, Role]:
        roles = dict()
        if not self._roles:
            logger.warning("No Managed RBAC Roles found, pushing new roles.")
            roles = self.ipf.settings.roles.create_roles(
                [
                    {"name": k, "description": v["description"], "policy_ids": [policies[_] for _ in v["policies"]]}
                    for k, v in self._roles.items()
                ]
            )
            return roles
        missing = {k: v for k, v in self._roles.items() if k not in self._current_roles}
        configured = {k: v for k, v in self._roles.items() if k in self._current_roles}
        if missing:
            logger.warning("Missing Managed RBAC Roles found, pushing new roles.")
            roles.update(
                self.ipf.settings.roles.create_roles(
                    [
                        {"name": k, "description": v["description"], "policy_ids": [policies[_] for _ in v["policies"]]}
                        for k, v in missing.items()
                    ]
                )
            )
        if configured:
            diff = DeepDiff(configured, self._current_roles, exclude_regex_paths=EXCLUDE_PATH, ignore_order=True)
            if diff:
                logger.warning("Role changes found, pushing updates.")
                for _ in diff.affected_root_keys.items:
                    logger.warning(f"Updating role: {_}.")
                    role = self.ipf.settings.roles.update_role(
                        role=self._current_roles[_]["role_id"],
                        policy_ids=[policies[p] for p in self._roles[_]["policies"]],
                        name=_,
                        description=self._roles[_]["description"],
                        replace=True,
                    )
                    roles.update({_: role})
                roles.update(
                    {
                        _: self.ipf.settings.roles.roles_by_name[_]
                        for _ in configured
                        if _ not in diff.affected_root_keys.items
                    }
                )
            else:
                roles.update({_: self.ipf.settings.roles.roles_by_name[_] for _ in configured})
        return roles

    def apply(self):
        policies = self._apply_policy()
        self.ipf.settings.roles.update()
        roles = self._apply_role(policies)
        return policies, roles



# TODO: REMOVE IN 8.0
import json
import logging
from collections import OrderedDict
from datetime import datetime
from os import path
from typing import Any, Union

from pydantic import Field
from pydantic.dataclasses import dataclass

from ipfabric.tools.shared import raise_for_status

logger = logging.getLogger("ipfabric")


@dataclass
class RestoreIntents:
    ipf: Any = Field(exclude=True)

    def __post_init__(self):
        logger.warning(
            "This is a deprecated class. Use the `IPFClient().settings.appliance_configuration` utility instead."
        )

    def restore_from_file(self, intents_file: str, groups_file: str, dashboard_file: str):
        if not all(path.exists(f) for f in [intents_file, groups_file, dashboard_file]):
            logger.critical("File(s) do not exist.")
            exit()
        with open(intents_file, "r") as file:
            intents = json.load(file)
        with open(groups_file, "r") as file:
            groups = json.load(file)
        with open(dashboard_file, "r") as file:
            dashboard = json.load(file)
        self._restore(intents, groups, dashboard)

    def restore_from_dictionary(self, intents: list, groups: list, dashboard: Union[dict, OrderedDict]):
        if not all(isinstance(f, list) for f in [intents, groups]) and not isinstance(dashboard, (dict, OrderedDict)):
            logger.critical("Intents and Groups must be Lists and Dashboard a Dictionary or OrderedDict.")
            exit()
        self._restore(intents, groups, dashboard)

    @staticmethod
    def _save_to_file(intents, groups, dashboard):
        tstamp = datetime.now().strftime("%Y%m%d-%H%M%S")

        def save(name, data):
            filename = name + "_" + tstamp + ".json"
            logger.info(f"Saving {name} to {filename}.")
            with open(filename, "w") as file:
                json.dump(data, file)
            logger.info(f"Saved {name} to {filename}.")

        save("intents", intents)
        save("groups", groups)
        save("dashboard", dashboard)

    def _restore(self, intents, groups, dashboard):
        i = input("\nDO YOU ACCEPT OVERRIDING ALL INTENTS, INTENT GROUPS, DASHBOARD SETTINGS? (y/n): ").strip().lower()
        if i and "y" not in i[0]:
            logger.warning("Exiting Script before making changes.")
            exit()

        old_intents = self.ipf.get("reports").json()
        old_intent_groups = self.ipf.get("reports/groups").json()
        old_dashboard = json.loads(self.ipf.get("settings/dashboard").text, object_pairs_hook=OrderedDict)
        self._save_to_file(old_intents, old_intent_groups, old_dashboard)

        logger.info("Deleting All Intents.")
        for i in old_intents:
            raise_for_status(self.ipf.delete(f"reports/{i['id']}"))
        logger.info("Deleted All Intents.")

        logger.info("Deleting All Intent Groups.")
        for g in old_intent_groups:
            raise_for_status(self.ipf.delete(f"reports/groups/{g['id']}"))
        logger.info("Deleted All Intent Groups.")

        intent_mapping = {i["id"]: None for i in intents}
        logger.info("Creating Intent Verification Rules.")
        for i in intents:
            i["groups"] = []
            i.pop("custom", None)
            src_id = i.pop("id", None)
            res = raise_for_status(self.ipf.post("reports", json=i))
            intent_mapping[src_id] = res.json()["id"]
        logger.info("Created Intent Verification Rules.")

        group_mapping = {g["id"]: None for g in groups}
        logger.info("Creating Intent Groups.")
        for g in groups:
            for c in g["children"]:
                c["id"] = intent_mapping[c["id"]]
            src_id = g.pop("id", None)
            g.pop("custom", None)
            res = raise_for_status(self.ipf.post("reports/groups", json=g))
            group_mapping[src_id] = res.json()["id"]
        logger.info("Created Intent Groups.")

        raw_dashboard = json.dumps(dashboard)
        logger.info("Creating Dashboard.")
        for src_group, dst_group in group_mapping.items():
            raw_dashboard = raw_dashboard.replace(src_group, dst_group)

        raise_for_status(self.ipf.put("settings/dashboard", json=json.loads(raw_dashboard)))
        logger.info("Created Dashboard.")
        return True



import logging
from typing import Any, Optional

from pydantic import BaseModel, Field, computed_field

from ipfabric.models.table import Table

logger = logging.getLogger("ipfabric")


class Fhrp(BaseModel):
    client: Any = Field(None, exclude=True)
    sn: Optional[str] = None

    def print_tables(self):
        print(sorted([_ for _ in dir(self) if _[0] != "_" and isinstance(getattr(self, _), Table)]))

    @computed_field
    @property
    def group_state(self) -> Table:
        return Table(client=self.client, endpoint="tables/fhrp/group-state", sn=self.sn)

    @computed_field
    @property
    def group_members(self) -> Table:
        return Table(client=self.client, endpoint="tables/fhrp/group-members", sn=self.sn)

    @computed_field
    @property
    def stproot_alignment(self) -> Table:
        return Table(client=self.client, endpoint="tables/fhrp/stproot-alignment", sn=self.sn)

    @computed_field
    @property
    def balancing(self) -> Table:
        return Table(client=self.client, endpoint="tables/fhrp/balancing", sn=self.sn)

    @computed_field
    @property
    def glbp_forwarders(self) -> Table:
        return Table(client=self.client, endpoint="tables/fhrp/glbp-forwarders", sn=self.sn)

    @computed_field
    @property
    def virtual_gateways(self) -> Table:
        return Table(client=self.client, endpoint="tables/fhrp/virtual-gateways", sn=self.sn)



import json
import logging
import re
from datetime import datetime
from typing import Optional, Any, Union, Literal, overload

from httpx import InvalidURL, URL
from pydantic import ConfigDict, BaseModel

from ipfabric.api import IPFabricAPI, check_deprecated
from ipfabric.diagrams import Diagram
from ipfabric.models import Technology, Inventory, Jobs, Intent, Devices, Extensions
from ipfabric.models.oas import Methods
from ipfabric.settings import Settings
from ipfabric.tools import TIMEZONES, raise_for_status, valid_snapshot

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None

logger = logging.getLogger("ipfabric")

RE_TABLE = re.compile(r"^tables/")
EXPORT_FORMAT = Literal["json", "csv", "df"]


class IPFClient(IPFabricAPI, BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    _inventory: Optional[Inventory] = None
    _technology: Optional[Technology] = None
    _jobs: Optional[Jobs] = None
    _diagram: Optional[Diagram] = None
    _intent: Optional[Intent] = None
    _devices: Optional[Devices] = None
    _settings: Optional[Settings] = None
    _last_snapshot_update: Optional[datetime] = None
    _extensions: Optional[Extensions] = None

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(self)
        self._inventory = Inventory(client=self)
        self._technology = Technology(client=self)
        self._jobs = Jobs(client=self)
        self._diagram = Diagram(ipf=self)
        self._intent = Intent(client=self)
        self._settings = Settings(client=self)
        self._extensions = Extensions(client=self)

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "timeout":
            self._client.timeout = value
        elif name == "snapshot_id":
            value = self._switch_snapshot(value)
            self.__dict__[name] = value
            if value and self.snapshot.disabled_intent_verification is False and self._intent:
                self._intent = Intent(client=self)
            if value and self._devices:
                self._devices = self.load_devices(device_attr_filters=self.attribute_filters)
            return None
        super().__setattr__(name, value)
        if name == "_attribute_filters" and self._devices:
            self._devices = self.load_devices(device_attr_filters=self.attribute_filters)

    def __repr__(self):
        return f"IPFClient(base_url='{str(self.base_url)}')"

    @property
    def inventory(self) -> Inventory:
        return self._inventory

    @property
    def technology(self) -> Technology:
        return self._technology

    @property
    def jobs(self) -> Jobs:
        return self._jobs

    @property
    def diagram(self) -> Diagram:
        return self._diagram

    @property
    def intent(self) -> Intent:
        return self._intent

    @property
    def settings(self) -> Settings:
        return self._settings

    @property
    def devices(self) -> Devices:
        """get devices"""
        if not self._devices:
            logger.info("Devices not loaded, loading devices.")
            self._devices = self.load_devices(device_attr_filters=self.attribute_filters)
        return self._devices

    @devices.setter
    def devices(self, devices: Devices):
        self._devices = devices

    @property
    def extensions(self) -> Extensions:
        return self._extensions

    def update(self) -> None:
        self._snapshots.update()
        if self.snapshot_id not in self.loaded_snapshots:
            logger.warning(f"Snapshot {self.snapshot_id} is no longer loaded switching to `$last`.")
            self.snapshot_id = "$last"

    def load_devices(self, device_filters: dict = None, device_attr_filters: dict = None) -> Devices:
        if device_attr_filters:
            logger.warning(
                f"Global Attribute Filter is applied, only returning devices matching: {self.attribute_filters}."
            )
        devices = Devices(
            client=self,
            snapshot_id=self.snapshot_id,
            device_filters=device_filters,
            device_attr_filters=device_attr_filters,
        )
        return devices

    def _create_payload(
        self,
        url: str,
        snapshot_id: Optional[str],
        filters: Optional[dict],
        sort: Optional[dict],
        attr_filters: Optional[dict],
    ) -> dict:
        """Optimized payload creation with reduced conditionals"""
        payload = {
            "format": {"dataType": "json"},
            "filters": filters or {},
            "sort": sort or {},
        }

        if snapshot_id:
            payload["snapshot"] = snapshot_id

        if RE_TABLE.match(url) and (attr_filters or self.attribute_filters):
            payload["attributeFilters"] = attr_filters or self.attribute_filters

        return payload

    def _check_url_payload(self, url, snapshot_id, filters, reports, sort, attr_filters, export, csv_tz):
        url = self._check_url(url)
        payload = self._create_payload(url, snapshot_id, filters, sort, attr_filters)
        oas_data = self.oas.get(url, Methods(full_api_endpoint="")).post

        if export != "csv":
            if isinstance(reports, (str, list)):
                payload["reports"] = reports
            elif reports is True and oas_data and oas_data.web_endpoint:
                payload["reports"] = oas_data.web_endpoint
            elif reports is True and (not oas_data or not oas_data.web_endpoint):
                logger.warning(
                    f"Could not automatically discover Web Endpoint for Intent Data for table '/{url}'.\n"
                    f"Returning results without Intent Rules."
                )
        else:
            if reports:
                logger.warning("CSV export does not return reports, parameter has been excluded.")
            payload["format"] = {"exportToFile": True, "dataType": "csv"}
            if csv_tz and csv_tz.lower() not in TIMEZONES:
                raise ValueError(
                    f"CSV timezone '{csv_tz}' not in available timezones; see `ipfabric.tools.shared.TIMEZONES`"
                )
            elif csv_tz:
                payload["format"]["options"] = {"timezone": TIMEZONES[csv_tz.lower()]}
        return url, payload

    def _get_payload(self, url, payload) -> Union[str, bool]:
        snapshot_id = payload.pop("snapshot", None)
        reports = payload.pop("reports") if "reports" in payload and isinstance(payload["reports"], str) else None
        p = "&".join([f"{k}={json.dumps(v, separators=(',', ':'))}" for k, v in payload.items()])
        if snapshot_id:
            p += f"&snapshot={snapshot_id}"
        if reports:
            p += f"&reports={reports}"
        url = self.base_url.join(url + f"?{p}")
        if len(str(url)) > 4096:
            return False
        return url

    @check_deprecated("post")
    def _stream(self, url, payload, export):
        get_url = self._get_payload(url, payload)
        if get_url is False and export == "csv":
            raise InvalidURL("URL exceeds max character limit of 4096 cannot export to CSV.")
        elif get_url is False:
            logger.warning("URL exceeds max character limit of 4096 switching to pagination.")
            return False
        with self.stream("GET", get_url) as stream_resp:
            data = stream_resp.read()
            raise_for_status(stream_resp)
        return data if export == "csv" else json.loads(data)["data"]

    def query(self, url: str, payload: Union[str, dict], get_all: bool = True) -> Union[list[dict], bytes]:
        """Submits a query, does no formatting on the parameters.  Use for copy/pasting from the webpage.

        Args:
            url: Example: https://demo1.ipfabric.io/api/v1/tables/vlan/device-summary or tables/vlan/device-summary
            payload: Dictionary to submit in POST or can be JSON string (i.e. read from file).
            get_all: Default use pager to get all results and ignore pagination information in the payload

        Returns:
            list or bytes: List of Dictionary objects or bytes if csv.
        """
        url = self._check_url(url)
        if isinstance(payload, str):
            payload = json.loads(payload)
        export = payload.get("format", {}).get("dataType", "json")
        data = False
        if export == "csv" or (self.streaming and get_all):
            data = self._stream(url, payload, export)
        elif get_all and data is False:
            data = list(self._ipf_pager(url, payload))
        elif data is False:
            res = raise_for_status(self.post(url, json=payload))
            data = res.json()["data"]
        return data

    def get_columns(self, url: str, ui: bool = False) -> list[str]:
        """Checks OAS to find available columns.

        Args:
            url: API url to post
            ui: True to return columns available in the UI, default False

        Returns:
            list: List of column names
        """
        ctype = "ui_columns" if ui else "columns"
        oas_data = getattr(self.oas.get(url, {}), "post", None)
        if oas_data and getattr(oas_data, ctype):
            return getattr(oas_data, ctype)
        raise LookupError(f"Could not find {'UI' if ui else 'API'} columns for table '/{url}'.")

    def get_count(
        self,
        url: str,
        filters: Optional[Union[dict, str]] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot_id: Optional[str] = None,
        snapshot: bool = True,
        reports: Optional[Union[bool, list, str]] = False,
    ) -> int:
        """Get a total number of rows
        Args:
            url: API URL to post to
            filters: Optional dictionary of filters
            attr_filters: Optional dictionary of attribute filters
            snapshot_id: Optional snapshot_id to override default
            snapshot: Set to False for some tables like management endpoints.
            reports: Boolean to return default reports, string of frontend URL where the reports are displayed,
                     or a list of report IDs
        Returns:
            int: a count of rows
        """
        snapshot_id = snapshot_id or self.snapshot_id if snapshot else None
        url, payload = self._fetch_setup(
            url, "json", None, snapshot_id, filters, reports, None, attr_filters, None, snapshot
        )
        payload.update({"columns": ["id"], "pagination": {"limit": 1, "start": 0}})
        res = raise_for_status(self.post(url, json=payload))
        return res.json()["_meta"]["count"]

    def _fetch_setup(self, url, export, columns, snapshot_id, filters, reports, sort, attr_filters, csv_tz, snapshot):
        if export == "df" and DataFrame is None:
            raise ImportError("pandas not installed. Run `pip install ipfabric[pd]`.")
        snapshot_id = snapshot_id or self.snapshot_id if snapshot else None
        if filters:
            filters = json.loads(filters) if isinstance(filters, str) else filters
        if "color" in json.dumps(filters) and not reports:
            reports = True
        url, payload = self._check_url_payload(url, snapshot_id, filters, reports, sort, attr_filters, export, csv_tz)
        payload["columns"] = columns or self.get_columns(url)
        cols = set(payload["columns"])
        oas_data = self.oas.get(url, Methods(full_api_endpoint="")).post
        if oas_data and cols.intersection(oas_data.deprecated_columns):
            logger.warning(
                f"API columns '{cols.intersection(oas_data.deprecated_columns)}' for endpoint "
                f"'/{oas_data.api_endpoint}' are marked deprecated in the OpenAPI specification."
            )
        return url, payload

    @overload
    def fetch(
        self,
        url,
        export: Literal["json"] = ...,
        columns: Optional[list[str]] = None,
        filters: Optional[Union[dict, str]] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot: bool = True,
    ) -> list[dict]: ...

    @overload
    def fetch(
        self,
        url,
        export: Literal["csv"],
        columns: Optional[list] = None,
        filters: Optional[Union[dict, str]] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot: bool = True,
        csv_tz: Optional[str] = None,
    ) -> bytes: ...

    @overload
    def fetch(
        self,
        url,
        export: Literal["df"] = ...,
        columns: Optional[list] = None,
        filters: Optional[Union[dict, str]] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot: bool = True,
    ) -> DataFrame: ...

    def fetch(
        self,
        url,
        export: EXPORT_FORMAT = "json",
        columns: Optional[list] = None,
        filters: Optional[Union[dict, str]] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot: bool = True,
        csv_tz: Optional[str] = None,
    ):
        """Gets data from IP Fabric for specified endpoint

        Args:
            url: Example tables/vlan/device-summary
            export: str: Export format to return [json, csv]; default is json.
            columns: Optional list of columns to return, None will return all
            filters: Optional dictionary of filters
            limit: Default to 1,000 rows
            start: Starts at 0
            snapshot_id: Optional snapshot_id to override default
            reports: Boolean to return default reports, string of frontend URL where the reports are displayed,
                     or a list of report IDs
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            attr_filters: Optional dictionary to apply an Attribute filter
            snapshot: Set to False for some tables like management endpoints.
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
        Returns:
            Union[list[dict], bytes, pandas.DataFrame]: List of dict if json, bytes string if CSV, DataFrame is df
        """
        url, payload = self._fetch_setup(
            url, export, columns, snapshot_id, filters, reports, sort, attr_filters, csv_tz, snapshot
        )
        payload["pagination"] = dict(start=start, limit=limit)
        data = False
        if export == "csv" or self.streaming:
            data = self._stream(url, payload, export)
        if data is False:
            res = raise_for_status(self.post(url, json=payload))
            data = res.json()["data"]
        if export == "df":
            data = DataFrame.from_records(data) if export == "df" else data
        return data

    @overload
    def fetch_all(
        self,
        url: str,
        export: Literal["json"] = ...,
        columns: Optional[list] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot: bool = True,
    ) -> list[dict]: ...

    @overload
    def fetch_all(
        self,
        url: str,
        export: Literal["csv"],
        columns: Optional[list] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot: bool = True,
        csv_tz: Optional[str] = None,
    ) -> bytes: ...

    @overload
    def fetch_all(
        self,
        url: str,
        export: Literal["df"],
        columns: Optional[list] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot: bool = True,
    ) -> DataFrame: ...

    def fetch_all(
        self,
        url: str,
        export: EXPORT_FORMAT = "json",
        columns: Optional[list] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        attr_filters: Optional[dict[str, list[str]]] = None,
        snapshot: bool = True,
        csv_tz: Optional[str] = None,
    ):
        """Gets all data from IP Fabric for specified endpoint

        Args:
            url: Example tables/vlan/device-summary
            export: str: Export format to return [json, csv]; default is json.
            columns: Optional list of columns to return, None will return all
            filters: Optional dictionary of filters
            snapshot_id: Optional snapshot_id to override default
            reports: Boolean to return default reports, string of frontend URL where the reports are displayed,
                     or a list of report IDs
            sort: Optional dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            attr_filters: Optional dictionary to apply an Attribute filter
            snapshot: Set to False for some tables like management endpoints.
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
        Returns:
            Union[list[dict], bytes, pandas.DataFrame]: List of dict if json, bytes string if CSV, DataFrame is df
        """
        url, payload = self._fetch_setup(
            url, export, columns, snapshot_id, filters, reports, sort, attr_filters, csv_tz, snapshot
        )
        data = False
        if export == "csv" or self.streaming:
            data = self._stream(url, payload, export)
        if data is False:
            data = list(self._ipf_pager(url, payload))
        if export == "df":
            data = DataFrame.from_records(data)
        return data

    def _shared_url(self, url: Union[int, str], table: bool = True) -> tuple[dict, str]:
        snapshot = None
        try:
            url_id = str(int(url))
        except ValueError:
            parsed_url = URL(url)
            snapshot = parsed_url.params.get("selectSnapshot", None)
            if table and parsed_url.params.get("copyId"):
                url_id = parsed_url.params.get("copyId")
            elif not table:
                url_id = parsed_url.path.split("/")[-1]
            else:
                raise SyntaxError("Wrong endpoint selected.")
        if table:
            resp = self.get(f"/tables/url/{url_id}")
        else:
            resp = self.get(f"/graphs/urls/{url_id}")
        return raise_for_status(resp).json(), snapshot

    def shared_view(self, url: Union[int, str], data=True, reports: bool = True) -> Union[str, list]:
        """Takes a shared table view link and returns the data or the code to implement in python.

        Args:
            url: Id of the shared view (1453653298) or full/partial URL (`/inventory/devices?copyId=1453653298`)
            data: Defaults to return the data instead of printing the code
            reports: True to return Intent Verification data

        Returns: List of dictionaries of the data or a string representing the code.
        """
        query, snapshot = self._shared_url(url, True)
        url = self.web_to_api[query["webEndpoint"]].api_endpoint
        hidden = {k for k, v in query.get("columnVisibility", {}).items() if not v}
        columns = {k for k, v in query.get("columnVisibility", {}).items() if v}
        columns.update({_ for _ in query.get("columnWidth", {}) if _ not in hidden})
        columns.update({_ for _ in self.get_columns(url, ui=True) if _ not in hidden})
        columns = sorted(list(columns))

        filters, sort = query.get("filters", {}), query.get("sort", {})
        filters.pop("selectedFilter", None)

        if data:
            return self.fetch_all(
                url=url,
                columns=columns,
                filters=filters,
                sort=sort,
                reports=query["webEndpoint"] if reports else False,
                snapshot_id=snapshot if snapshot else self.snapshot_id,
            )
        code = f"{url.split('/')[-1]} = ipf.fetch_all(\n    {url=},\n    {columns=},"
        code += f"\n    {filters=}," if filters else ""
        code += f"\n    {sort=}," if sort else ""
        code += f"\n    reports='{query['webEndpoint']}'," if reports else ""
        code += f"\n    snapshot_id='{snapshot}'," if snapshot else ""
        code += "\n)"
        return code

    def share_link(  # noqa: C901
        self,
        url: str,
        filters: Union[dict, str],
        columns: Optional[list] = None,
        sort: Optional[dict] = None,
        snapshot: Optional[Union[bool, str]] = False,
    ) -> str:
        """Creates a shareable link for a table.

        Args:
            url: Example tables/vlan/device-summary
            filters: Filter dictionary to create view for
            columns: Optional list of columns to display, defaults to all. Some columns are hidden in the UI and will
                     not be displayed.
            sort: Dict of column to sort on
            snapshot: Default False do not add selector, True use IPFClient snapshot, string specific snapshot.
        Returns:
            URL string
        """
        url = self._check_url(url)
        try:
            web_endpoint = self.oas[url].post.web_endpoint
        except KeyError:
            raise NotImplementedError(f"Cannot create Shared View URL for table `{url}` as it has no web endpoint.")
        _columns = {_: True for _ in self.get_columns(url)}

        if columns:
            _columns.update({_: False for _ in _columns if _ not in columns})
        try:
            hidden = list(
                {k for k, v in _columns.items() if v and k != "id"}.difference(set(self.get_columns(url, ui=True)))
            )
            if hidden:
                logger.warning(
                    f"Column(s) `{hidden}` are hidden in the UI and will not be displayed using the shared view link."
                )
        except LookupError:
            pass
        try:
            user = self.get("tables/settings", params={"name": web_endpoint}).json()["userSettings"]
            auto_size, col_width = user.get("autoSizedColumns", True), user.get("columnWidth", {})
        except KeyError:
            auto_size, col_width = True, {}
        payload = {
            "autoSizedColumns": auto_size,
            "columnVisibility": _columns,
            "webEndpoint": web_endpoint,
            "filters": json.loads(filters) if isinstance(filters, str) else filters,
        }
        payload.update({"columnWidth": col_width} if col_width else {})
        payload.update({"sort": sort} if sort else {})
        resp = raise_for_status(self.post("tables/url", json=payload))
        if not snapshot:
            select_snap = ""
        elif snapshot is True:
            select_snap = f"&selectSnapshot={self.snapshot_id}"
        elif valid_snapshot(snapshot) and snapshot in self.snapshots:
            select_snap = f"&selectSnapshot={snapshot}"
        return str(self.base_url.join(web_endpoint)) + f"?copyId={resp.json()['id']}" + select_snap

    def post_to_get(self, url: str, payload: Union[dict, str], ignore_reports: bool = False):
        """

        Args:
            url: str: API or Frontend URL
            payload: Union[str, dict]: Copy/Paste payload from API description in UI, only required object is 'snapshot'
            ignore_reports: bool: Remove the reports from the payload (intent data) from being included.

        Returns: str: Get URL to return table data instead of pagination

        """
        if isinstance(payload, str):
            payload = json.loads(payload)
        snapshot = payload.get("snapshot", None)
        url, payload = self._fetch_setup(
            url,
            "json",
            payload.get("columns", None),
            snapshot,
            payload.get("filters", None),
            None if ignore_reports else payload.get("reports", None),
            payload.get("sort", None),
            payload.get("attributeFilters", None),
            None,
            True if snapshot else False,
        )
        return str(self._get_payload(url, payload))



import argparse
import os
from collections import defaultdict
from pathlib import Path

from ipfabric import IPFClient
from ipfabric.models import Devices
from ipfabric.scripts.shared import load_env, base_args, parse_base_args
from ipfabric.tools.shared import Vulnerabilities

try:
    import pandas as pd
    import openpyxl  # noqa: F401
except ImportError:
    raise ImportError("pandas and openpyxl are required, please install by using `pip3 install pandas openpyxl`.")


def load_cve(ipf: IPFClient, nvd_api_key: str, timeout=60, retries: int = 2):
    vuln = Vulnerabilities(ipf=ipf, nvd_api_key=nvd_api_key, timeout=timeout, retries=retries)
    versions = vuln.check_versions()
    cves = defaultdict(dict)
    for v in versions:
        if v.family not in cves[v.vendor]:
            cves[v.vendor][v.family] = defaultdict(dict)
        cves[v.vendor][v.family][v.version] = v.cves
    return cves


def load_eol(ipf: IPFClient) -> pd.DataFrame:
    eol = ipf.fetch_all(
        "tables/reports/eof/detail",
        columns=[
            "hostname",
            "siteName",
            "deviceSn",
            "vendor",
            "pid",
            "replacement",
            "sn",
            "endSale",
            "endMaintenance",
            "endSupport",
            "url",
            "dscr",
        ],
        filters={
            "or": [
                {"endSale": ["empty", False]},
                {"endMaintenance": ["empty", False]},
                {"endSupport": ["empty", False]},
            ]
        },
    )
    if eol:
        df_eol = pd.DataFrame(eol)
        for eolt in ["endSale", "endMaintenance", "endSupport"]:
            df_eol[eolt] = df_eol[eolt].values.astype(dtype="datetime64[ms]")
    else:
        df_eol = pd.DataFrame()
    return df_eol


def _check_cves(dev, cves, data, clean, error):
    for cve in cves:
        if cve.error:
            error.append(
                [
                    dev.hostname,
                    dev.site_name,
                    dev.sn_hw,
                    str(dev.login_ip),
                    dev.vendor,
                    dev.family,
                    dev.version,
                    None,
                    cve.error,
                ]
            )
            continue
        elif cve.total_results == 0:
            clean.append(
                [
                    dev.hostname,
                    dev.site_name,
                    dev.sn_hw,
                    str(dev.login_ip),
                    dev.vendor,
                    dev.family,
                    dev.version,
                    cve.cpe.cpeName,
                ]
            )
            continue
        for c in cve.cves:
            data.append(
                [
                    dev.hostname,
                    dev.site_name,
                    dev.sn_hw,
                    str(dev.login_ip),
                    dev.vendor,
                    dev.family,
                    dev.version,
                    cve.cpe.cpeName,
                    c.cve_id,
                    c.base_score,
                    c.description,
                    c.url,
                ]
            )
    return data, clean, error


def parse_cve(devices: Devices, cves):
    data, clean, error = list(), list(), list()
    for dev in devices.all:
        if (
            dev.vendor
            and dev.version
            and dev.vendor in cves
            and dev.family in cves[dev.vendor]
            and dev.version in cves[dev.vendor][dev.family]
        ):
            data, clean, error = _check_cves(dev, cves[dev.vendor][dev.family][dev.version], data, clean, error)
        else:
            error.append(
                [
                    dev.hostname,
                    dev.site_name,
                    dev.sn_hw,
                    str(dev.login_ip),
                    dev.vendor,
                    dev.family,
                    dev.version,
                    None,
                    "Unknown Error",
                ]
            )
    return (
        pd.DataFrame(
            data,
            columns=[
                "device",
                "site",
                "serial",
                "ip",
                "vendor",
                "family",
                "version",
                "cpe_name",
                "cve_id",
                "cve_base_score",
                "cve_description",
                "cve_url",
            ],
        ),
        pd.DataFrame(
            clean,
            columns=[
                "device",
                "site",
                "serial",
                "ip",
                "vendor",
                "family",
                "version",
                "cpe_name",
            ],
        ),
        pd.DataFrame(
            error, columns=["device", "site", "serial", "ip", "vendor", "family", "version", "cpe_name", "error"]
        ),
    )


def write_json(df_eol: pd.DataFrame, df_cve: pd.DataFrame, df_clean: pd.DataFrame, df_error: pd.DataFrame, filename):
    json_data = (
        f"{{"
        f'"CVE": {df_cve.to_json(orient="records", date_format="iso")}, '
        f'"No CVE": {df_clean.to_json(orient="records", date_format="iso")}, '
        f'"CVE Error": {df_error.to_json(orient="records", date_format="iso")}, '
        f'"EoL": {df_eol.to_json(orient="records", date_format="iso")}'
        f"}}"
    )
    with open(filename, "w") as f:
        f.write(json_data)


def write_excel(df_eol: pd.DataFrame, df_cve: pd.DataFrame, df_clean: pd.DataFrame, df_error: pd.DataFrame, filename):
    writer = pd.ExcelWriter(filename, engine="openpyxl")
    df_eol.to_excel(writer, sheet_name="End of Life", index=False)
    df_cve.to_excel(writer, sheet_name="CVE DATA", index=False)
    df_clean.to_excel(writer, sheet_name="Clean Devices", index=False)
    df_error.to_excel(writer, sheet_name="CVE ERROR", index=False)
    writer.close()


def main():
    load_env()
    arg_parser = argparse.ArgumentParser(
        description="Creates an Excel report of CVEs and EoL for devices in IP Fabric; requires:\n"
        "Python Packages: 'pandas' and 'openpyxl' "
        "(`pip install ipfabric[cve]` or `pip install pandas openpyxl`)\n"
        "NVD_API_KEY: See https://nvd.nist.gov/developers/request-an-api-key",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    arg_parser = base_args(arg_parser)
    arg_parser.add_argument(
        "-n",
        "--nvd-api-key",
        help="NVD_API_KEY from https://nvd.nist.gov/developers/request-an-api-key",
    )
    arg_parser.add_argument(
        "-f",
        "--filename",
        help="Filename to save, defaults to CVE_report.(xlsx|json). "
        "Extension will always be corrected to either '.xlsx' or '.json'.",
    )
    arg_parser.add_argument(
        "-to", "--timeout", help="Timeout for NIST API (Int, Float, or None); default 60 seconds.", default=60
    )
    arg_parser.add_argument("-r", "--retries", help="Number of retries for NIST API errors; default 2.", default=2)
    arg_parser.add_argument(
        "-l", "--site-name", help="Filter on a specific site name (location); case-insensitive equals search."
    )
    arg_parser.add_argument(
        "-j",
        "--json",
        help="Save as JSON output instead of Excel; default is Excel output.",
        action="store_true",
        default=False,
    )
    args = arg_parser.parse_args()
    args = parse_base_args(args)
    if isinstance(args.timeout, str) and args.timeout.lower() == "none":
        args.timeout = None
    else:
        args.timeout = float(args.timeout)
    if args.filename:
        args.filename = Path(args.filename).resolve().with_suffix(".json" if args.json else ".xlsx").absolute()

    ipf = IPFClient(snapshot_id=args.snapshot, base_url=args.base_url, auth=args.auth)

    if args.site_name:
        site = ipf.inventory.sites.all(filters={"siteName": ["ieq", args.site_name]})
        if not site:
            raise ValueError(f"Site Name {args.site_name} not found.")
        elif len(site) > 1:
            raise ValueError(f"Multiple Site Names found for {args.site_name}.")
        ipf.attribute_filters = {"siteName": [site[0]["siteName"]]}

    nvd_api_key = args.nvd_api_key or getattr(ipf, "nvd_api_key", None) or os.environ.get("NVD_API_KEY", None)
    if not nvd_api_key:
        raise ValueError("NIST NVD_API_KEY parsed nor defined in environment variables.")

    df_cve, df_clean, df_error = parse_cve(ipf.devices, load_cve(ipf, nvd_api_key, args.timeout, int(args.retries)))
    df_eol = load_eol(ipf)

    if args.json:
        write_json(df_eol, df_cve, df_clean, df_error, args.filename or "CVE_report.json")
    else:
        write_excel(df_eol, df_cve, df_clean, df_error, args.filename or "CVE_report.xlsx")


if __name__ == "__main__":
    main()



import json
import logging
from typing import Union, Optional, Literal, Any

from pydantic import BaseModel

from ipfabric.models import Device
from ipfabric.models.security import DefaultAction
from ipfabric.tools.shared import raise_for_status
from .input_models import (
    Unicast,
    Multicast,
    Host2GW,
    Network,
    NetworkSettings,
    PathLookupSettings,
    Overlay,
    SharedView,
    GroupSettings,
)
from .output_models.graph_result import (
    GraphResult,
    Position,
    PathLookupResult,
    TopologyResult,
)

logger = logging.getLogger("ipfabric")
GRAPHS_URL = "graphs/"
GRAPH_TYPES = Union[Unicast, Multicast, Host2GW, Network]
GRAPH_SETTINGS = Optional[Union[NetworkSettings, PathLookupSettings]]
OVERLAY_SETTINGS = Optional[Union[Overlay, dict]]
ATTRIBUTE_FILTERS = Optional[dict[str, list[str]]]
POSITION_SETTINGS = Optional[dict[str, Union[Position, dict]]]
GRAPH_RESULTS = Union[PathLookupResult, TopologyResult]


class GraphDevice(BaseModel):
    device: Device
    decidingPolicyName: Optional[str] = None
    decidingRule: Optional[list] = None
    securityType: Optional[str] = None
    securityModelRule: Optional[Union[list, DefaultAction]] = None


class Diagram(BaseModel):
    ipf: Any

    def _check_snapshot_id(self, snapshot_id):
        snapshot_id = snapshot_id or self.ipf.snapshot_id
        if snapshot_id not in self.ipf.loaded_snapshots:
            raise ValueError(f"Snapshot {snapshot_id} is not loaded or not found in IP Fabric.")
        return snapshot_id

    def _intent_overlay(self, overlay: Overlay) -> dict:
        try:
            Overlay.valid_intent_rule(overlay.intentRuleId)
        except ValueError:
            if not self.ipf.intent.loaded:
                self.ipf.intent.load_intent()
            intents = self.ipf.intent.intents_by_name[overlay.intentRuleId]
            if len(intents) > 1:
                raise ValueError(f"Multiple Intents found with name `{overlay.intentRuleId}`.")
            else:
                overlay.intentRuleId = intents[0].intent_id
        return overlay.model_dump()

    def _snapshot_overlay(self, overlay: Overlay, snapshot_id: str = None) -> dict:
        if overlay.snapshotToCompare not in self.ipf.loaded_snapshots:
            raise ValueError(f"Snapshot `{overlay.snapshotToCompare}` is not loaded.")
        overlay.snapshotToCompare = self.ipf.snapshots[overlay.snapshotToCompare].snapshot_id
        if snapshot_id or self.ipf.snapshot_id == overlay.snapshotToCompare:
            raise ValueError(f"Cannot compare snapshot `{overlay.snapshotToCompare}` to itself.")
        self._check_snapshot_id(overlay.snapshotToCompare)
        return overlay.model_dump()

    def _format_overlay(self, overlay: Union[Overlay, dict], snapshot_id: str = None) -> dict:
        if isinstance(overlay, dict):
            overlay = Overlay(**overlay)
        if overlay.intentRuleId:
            return self._intent_overlay(overlay)
        return self._snapshot_overlay(overlay, snapshot_id)

    def _query(
        self,
        parameters: Union[GRAPH_TYPES, dict],
        snapshot_id: Optional[str] = None,
        overlay: OVERLAY_SETTINGS = None,
        image: Literal["png", "svg", "json", "vsdx"] = "json",
        graph_settings: Union[GRAPH_SETTINGS, dict] = None,
        attr_filters: ATTRIBUTE_FILTERS = None,
        positions: POSITION_SETTINGS = None,
    ) -> Union[dict, bytes]:
        """
        Submits a query, does no formatting on the parameters.  Use for copy/pasting from the webpage.
        :param parameters: dict: Dictionary to submit in POST.
        :return: list: List of Dictionary objects.
        """
        url = GRAPHS_URL
        if image in ["svg", "png", "vsdx"]:
            url += image
        payload = dict(
            parameters=parameters if isinstance(parameters, dict) else parameters.model_dump(),
            snapshot=self._check_snapshot_id(snapshot_id),
        )
        if overlay:
            payload["overlay"] = self._format_overlay(overlay, snapshot_id)
        if graph_settings:
            payload["settings"] = graph_settings if isinstance(graph_settings, dict) else graph_settings.model_dump()
        if attr_filters or self.ipf.attribute_filters:
            payload["attributeFilters"] = attr_filters or self.ipf.attribute_filters
        payload["positions"] = {k: dict(v) for k, v in positions.items()} if positions else {}
        if image == "vsdx":
            if payload.get("overlay", None):
                logger.warning("Overlay is not supported for Visio diagrams.")
            json_graph = raise_for_status(self.ipf.post(GRAPHS_URL, json=payload)).json()["graphResult"]
            files = {
                "graph": (
                    "graph.json",
                    json.dumps(
                        {
                            "graph": {
                                "boxLabels": json_graph["boxLabels"],
                                "edges": json_graph["graphData"]["edges"],
                                "nodes": json_graph["graphData"]["nodes"],
                            }
                        }
                    ),
                    "application/json",
                )
            }
            res = raise_for_status(self.ipf.post(url, files=files))
        else:
            res = raise_for_status(self.ipf.post(url, json=payload))
        return res.json() if image == "json" else res.content

    @staticmethod
    def _swap_src_dst(p: GRAPH_TYPES, unicast_swap_src_dst: bool = False) -> GRAPH_TYPES:
        return p.swap_src_dst if unicast_swap_src_dst and isinstance(p, Unicast) else p

    def json(
        self,
        parameters: GRAPH_TYPES,
        snapshot_id: Optional[str] = None,
        overlay: OVERLAY_SETTINGS = None,
        graph_settings: GRAPH_SETTINGS = None,
        attr_filters: ATTRIBUTE_FILTERS = None,
        unicast_swap_src_dst: bool = False,
        positions: POSITION_SETTINGS = None,
    ) -> dict:
        return self._query(
            self._swap_src_dst(parameters, unicast_swap_src_dst),
            snapshot_id=snapshot_id,
            image="json",
            overlay=overlay,
            attr_filters=attr_filters,
            graph_settings=graph_settings.model_dump() if graph_settings else None,
            positions=positions,
        )

    def svg(
        self,
        parameters: GRAPH_TYPES,
        snapshot_id: Optional[str] = None,
        overlay: OVERLAY_SETTINGS = None,
        graph_settings: GRAPH_SETTINGS = None,
        attr_filters: ATTRIBUTE_FILTERS = None,
        unicast_swap_src_dst: bool = False,
        positions: POSITION_SETTINGS = None,
    ) -> bytes:
        return self._query(
            self._swap_src_dst(parameters, unicast_swap_src_dst),
            snapshot_id=snapshot_id,
            overlay=overlay,
            attr_filters=attr_filters,
            image="svg",
            graph_settings=graph_settings.model_dump() if graph_settings else None,
            positions=positions,
        )

    def png(
        self,
        parameters: GRAPH_TYPES,
        snapshot_id: Optional[str] = None,
        overlay: OVERLAY_SETTINGS = None,
        graph_settings: GRAPH_SETTINGS = None,
        attr_filters: ATTRIBUTE_FILTERS = None,
        unicast_swap_src_dst: bool = False,
        positions: POSITION_SETTINGS = None,
    ) -> bytes:
        return self._query(
            self._swap_src_dst(parameters, unicast_swap_src_dst),
            snapshot_id=snapshot_id,
            overlay=overlay,
            attr_filters=attr_filters,
            image="png",
            graph_settings=graph_settings.model_dump() if graph_settings else None,
            positions=positions,
        )

    def visio(
        self,
        parameters: GRAPH_TYPES,
        snapshot_id: Optional[str] = None,
        overlay: OVERLAY_SETTINGS = None,
        graph_settings: GRAPH_SETTINGS = None,
        attr_filters: ATTRIBUTE_FILTERS = None,
        unicast_swap_src_dst: bool = False,
        positions: POSITION_SETTINGS = None,
    ) -> bytes:
        return self._query(
            self._swap_src_dst(parameters, unicast_swap_src_dst),
            snapshot_id=snapshot_id,
            overlay=overlay,
            attr_filters=attr_filters,
            image="vsdx",
            graph_settings=graph_settings.model_dump() if graph_settings else None,
            positions=positions,
        )

    def share_link(
        self,
        parameters: GRAPH_TYPES,
        snapshot_id: Optional[str] = None,
        overlay: OVERLAY_SETTINGS = None,
        graph_settings: GRAPH_SETTINGS = None,
        attr_filters: ATTRIBUTE_FILTERS = None,
        unicast_swap_src_dst: bool = False,
        positions: POSITION_SETTINGS = None,
    ) -> str:
        parameters = self._swap_src_dst(parameters, unicast_swap_src_dst)
        resp = self._query(
            parameters,
            snapshot_id=snapshot_id,
            overlay=overlay,
            image="json",
            attr_filters=attr_filters,
            graph_settings=graph_settings.model_dump() if graph_settings else None,
            positions=positions,
        )

        input_model = parameters.model_dump()
        input_model.pop("layouts", None)
        payload = {
            "graphView": {
                "name": "Shared view",
                "parameters": input_model,
                "collapsedNodeGroups": [],
                "hiddenNodes": [],
                "positions": {k: v["position"] for k, v in resp["graphResult"]["graphData"]["nodes"].items()},
                "settings": resp["graphResult"]["settings"],
            },
            "snapshot": self._check_snapshot_id(snapshot_id),
        }
        if overlay:
            payload["graphView"]["overlay"] = self._format_overlay(overlay, snapshot_id)
        res = raise_for_status(self.ipf.post("graphs/urls", json=payload))
        return str(self.ipf.base_url.join(f"/diagrams/share/{res.json()['id']}"))

    def model(
        self,
        parameters: GRAPH_TYPES,
        snapshot_id: Optional[str] = None,
        overlay: OVERLAY_SETTINGS = None,
        graph_settings: GRAPH_SETTINGS = None,
        attr_filters: ATTRIBUTE_FILTERS = None,
        unicast_swap_src_dst: bool = False,
        positions: POSITION_SETTINGS = None,
    ) -> GRAPH_RESULTS:
        json_data = self.json(
            parameters, snapshot_id, overlay, graph_settings, attr_filters, unicast_swap_src_dst, positions
        )
        graph_result = TopologyResult(**json_data) if "topology" in json_data else PathLookupResult(**json_data)
        e_setting = self._diagram_edge_settings(graph_result.graphResult.settings)

        edges = {edge_id: edge for edge_id, edge in graph_result.edges.items()}
        for edge_id, edge in edges.items():
            edge.protocol = e_setting[edge.edgeSettingsId].name if edge.edgeSettingsId in e_setting else None
            if edge.source:
                edge.source = graph_result.nodes[edge.source]
            if edge.target:
                edge.target = graph_result.nodes[edge.target]

        graph_result.graphResult.graphData.edges = edges
        return graph_result if isinstance(graph_result, TopologyResult) else self._diagram_pathlookup(graph_result)

    @staticmethod
    def _diagram_pathlookup(graph_result: PathLookupResult) -> PathLookupResult:
        for edge_id, edge in graph_result.edges.items():
            for prev_id in edge.prevEdgeIds:
                edge.prevEdge.append(graph_result.edges[prev_id])
            for next_id in edge.nextEdgeIds:
                edge.nextEdge.append(graph_result.edges[next_id] if next_id in graph_result.edges else next_id)
        return graph_result

    @staticmethod
    def _diagram_edge_settings(graph_settings: Union[NetworkSettings, PathLookupSettings]) -> dict:
        edge_setting_dict = dict()
        for edge in graph_settings.edges:
            edge_setting_dict[edge.id] = edge
            if isinstance(edge, GroupSettings):
                for child in edge.children:
                    edge_setting_dict[child.id] = child
        return edge_setting_dict

    def shared_view(
        self,
        url: Union[int, str],
        image: Literal["json", "code", "model", "svg", "png"] = "json",
        positions: bool = False,
    ):
        """Takes a shared graph link and returns the data or the code to implement in python.

        Args:
            url: Id of the shared view (1453653298) or full/partial URL (`/diagrams/share/1453626097`)
            image: Defaults to return the data instead of printing the code
            positions: If returning code then include positions of nodes in example; Default False

        Returns: The graph data or string representing the code to produce it.
        """
        query, _ = self.ipf._shared_url(url, False)
        if query["snapshot"] not in self.ipf.loaded_snapshots:
            logger.warning(f'Snapshot {query["snapshot"]} is not loaded, switching to {self.ipf.snapshot_id}.')
            query["snapshot"] = self.ipf.snapshot_id
        overlay = query["graphView"].get("overlay", {})
        if (
            overlay
            and overlay.get("snapshotToCompare", None)
            and overlay["snapshotToCompare"] not in self.ipf.loaded_snapshots
        ):
            logger.error(f"Snapshot `{overlay['snapshotToCompare']}` is not loaded to compare, removing overlay.")
            overlay = {}

        view = SharedView(
            snapshot_id=query["snapshot"],
            client_snapshot_id=self.ipf.snapshot_id,
            hidden_nodes=query["graphView"]["hiddenNodes"],
            collapsed_node_groups=query["graphView"]["collapsedNodeGroups"],
            positions=query["graphView"]["positions"],
            settings=query["graphView"]["settings"]["edges"],
            params=query["graphView"]["parameters"],
            hidden_devs=query["graphView"]["settings"].get("hiddenDeviceTypes", []),
            path_lookup=query["graphView"]["settings"].get("pathLookup", None),
            overlay=overlay,
        )
        if view.hidden_nodes or view.collapsed_node_groups:
            logger.warning("Hidden Nodes and Collapsed Node Groups are only available via the UI.")

        if image in ["json", "svg", "png"]:
            return self._query(
                parameters=query["graphView"]["parameters"],
                snapshot_id=query["snapshot"],
                overlay=overlay,
                image=image,
                graph_settings=query["graphView"]["settings"],
                positions=query["graphView"]["positions"],
            )

        if image == "code":
            return "\n".join(view.create_code(positions))
        if image == "model":
            return self.model(
                parameters=view.graph_model(),
                snapshot_id=query["snapshot"],
                overlay=overlay,
                graph_settings=view.graph_settings(),
                positions=query["graphView"]["positions"],
            )
        return None

    def _process_path(
        self, path: Union[GRAPH_RESULTS, GRAPH_TYPES], snapshot_id: Optional[str] = None
    ) -> GRAPH_RESULTS:
        """Helper method to process model or return a GraphResult.

        Args:
            path: PathLookupResult or TopologyResult or Unicast, Multicast, Host2GW, Network object
            snapshot_id: Optional snapshot_id to use

        Returns:
            GraphResult: Processed result
        """
        if isinstance(path, (PathLookupResult, TopologyResult)):
            return path
        return self.model(path, snapshot_id=snapshot_id)

    def nodes_in_diagram(
        self,
        path: Union[GRAPH_RESULTS, GRAPH_TYPES],
        dev_types: Optional[list[str]] = None,
        inventory_columns: Optional[list[str]] = None,
        snapshot_id: Optional[str] = None,
    ) -> list[Device]:
        """Fetches the device inventory for devices in the path.

        Args:
            path: PathLookupResult or TopologyResult or Unicast, Multicast, Host2GW, Network object
            dev_types: Optional list of device types to fetch, default ["l3switch", "fw", "switch", "router", "lb"]
            inventory_columns: Optional list of columns to fetch, default
                               ["sn", "version", "hostname", "model", "vendor", "siteName"]
            snapshot_id: Optional snapshot_id to use
        Returns:
            list[dict]: Device inventory
        """
        list_of_device_objects = list()
        dev_types = NetworkSettings._valid_dev_types(dev_types or ["l3switch", "fw", "switch", "router", "lb"])
        inventory_columns = inventory_columns or ["sn", "version", "hostname", "model", "vendor", "siteName"]
        if not all(_ in self.ipf.oas["tables/inventory/devices"].post.columns for _ in inventory_columns):
            raise ValueError(
                f"Inventory Columns '{inventory_columns}' must be in "
                f"{self.ipf.oas['tables/inventory/devices'].post.columns}."
            )

        path = self._process_path(path, snapshot_id)
        device_sn_to_fetch = [node.sn for node in path.nodes.values() if node.type in dev_types]
        for sn in device_sn_to_fetch:
            list_of_device_objects.append(self.ipf.devices.by_sn[sn])
        return list_of_device_objects

    def security_event_rules(
        self,
        path: Union[GraphResult, Unicast, Multicast, Host2GW],
        snapshot_id: Optional[str] = None,
    ) -> list[GraphDevice]:  # noqa: C901
        """Fetches the security events for devices in the path.

        Args:
            path: PathLookupResult or TopologyResult or Unicast, Multicast, Host2GW
            snapshot_id: Optional snapshot_id to use

        Returns:
            list[dict]: Device Inventory Information and Security Event
        """
        logger.warning("This method is experimental and may not work as expected.")
        path = self._process_path(path, snapshot_id)
        app_graph_nodes_dict = {node.id: node for node in path.nodes.values() if node.style}
        path_inventory_dict = {device.sn: device for device in self.nodes_in_diagram(path, snapshot_id)}
        list_for_return = list()

        for vdevice, decisions in path.pathlookup.decisions.items():
            node = app_graph_nodes_dict.get(vdevice, None)
            if not node or not path_inventory_dict.get(node.sn, None):
                continue

            for trace in decisions.traces:
                for chain in trace.trace:
                    for event in chain.events:
                        if event.type in ["security check", "security check ignored"]:
                            device_data = path_inventory_dict[node.sn]
                            graph_device = GraphDevice(
                                device=device_data,
                                decidingPolicyName=event.decidingPolicyName,
                                decidingRule=event.decidingRule,
                                securityType=event.securityType,
                            )
                            if event.securityType == "zoneFw":
                                sec_model = graph_device.device.get_security_model()
                                machine_zones = sec_model.security.machineZones
                                rule_sequence = event.decidingRule[1]
                                zone_firewall_rules = (
                                    graph_device.device.technology.security.zone_firewall_policies.all(
                                        filters={
                                            "policyName": ["like", event.decidingPolicyName],
                                            "ruleNameNumeric": ["eq", (rule_sequence + 1)],
                                        }
                                    )
                                )
                                if zone_firewall_rules:
                                    graph_device.securityModelRule = zone_firewall_rules[0]
                                else:
                                    graph_device.securityModelRule = machine_zones.ruleChains.get(
                                        event.decidingPolicyName
                                    )
                            list_for_return.append(graph_device)
        return list_for_return

    def stp_instances(self, sites: Union[list[str], str], vlan_id: int) -> list[dict]:
        """Get the STP instances for the given sites and VLAN ID.

        Args:
            sites: Site name or list of site names to search for.
            vlan_id: VLAN ID to search for.

        Returns:

        """
        sites = [sites] if isinstance(sites, str) else sites
        filters = {"or": [{"siteName": ["eq", _]} for _ in sites]}
        devs = [_["sn"] for _ in self.ipf.inventory.devices.all(filters=filters, columns=["sn"])]
        roots = self.ipf.technology.stp.instance_members.all(
            filters={
                "and": [
                    {"members": ["containsAny", devs]},
                    {"rootHostname": ["empty", False]},
                    {"vlanId": ["eq", vlan_id]},
                ]
            },
            columns=["rootHostname", "rootId", "vlanId", "vlanName"],
        )
        return roots



import logging
from typing import Any, Optional

from pydantic import BaseModel, Field, computed_field

from ipfabric.models.table import Table

logger = logging.getLogger("ipfabric")


class Sdwan(BaseModel):
    client: Any = Field(None, exclude=True)
    sn: Optional[str] = None

    def print_tables(self):
        print(sorted([_ for _ in dir(self) if _[0] != "_" and isinstance(getattr(self, _), Table)]))

    @computed_field
    @property
    def versa_sites(self) -> Table:
        return Table(client=self.client, endpoint="tables/sdwan/versa/sites", sn=self.sn)

    @computed_field
    @property
    def versa_links(self) -> Table:
        return Table(client=self.client, endpoint="tables/sdwan/versa/links", sn=self.sn)

    @computed_field
    @property
    def silverpeak_overlay(self) -> Table:
        return Table(client=self.client, endpoint="tables/sdwan/silverpeak/overlay", sn=self.sn)

    @computed_field
    @property
    def silverpeak_underlay(self) -> Table:
        return Table(client=self.client, endpoint="tables/sdwan/silverpeak/underlay", sn=self.sn)

    @computed_field
    @property
    def viptela_bfd_sessions(self) -> Table:
        return Table(client=self.client, endpoint="tables/sdwan/viptela/bfd-sessions", sn=self.sn)

    @computed_field
    @property
    def viptela_bfd_summary(self) -> Table:
        return Table(client=self.client, endpoint="tables/sdwan/viptela/bfd-summary", sn=self.sn)

    @computed_field
    @property
    def viptela_control_connections(self) -> Table:
        return Table(client=self.client, endpoint="tables/sdwan/viptela/control-connections", sn=self.sn)

    @computed_field
    @property
    def velocloud_overlay(self) -> Table:
        return Table(client=self.client, endpoint="tables/sdwan/velocloud/overlay", sn=self.sn)



"""
Python3 script to search for an IPv4, IPv6, or MAC address in multiple IP Fabric's tables and prints the output.
"""

import argparse
import json
import logging
import re
from copy import deepcopy
from typing import Union

from ipfabric import IPFClient
from ipfabric.models.global_search import GlobalSearch, RANKED
from ipfabric.scripts.shared import shared_args

try:
    from rich.console import Console

    CONSOLE = Console()
except ImportError:
    CONSOLE = None

logging.basicConfig(format="%(levelname)s: %(message)s")
LOGGER = logging.getLogger("ipf_global_search")


def main() -> dict[str, dict[str, Union[str, list]]]:
    arg_parser = argparse.ArgumentParser(
        description="Search all tables for an IPv4, IPv6, or MAC address. "
        "This does not include IPv4 or IPv6 Route tables as too many results would be returned.",
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=f"""
Tables are searched in the order as listed below.  If Full Scan option is enabled then any other tables are
printed after these results in no specific order.\n

Default Tables for IPv4 Search:
{chr(10).join([f'{idx + 1}. {_}' for idx, _ in enumerate(RANKED['ipv4'])])}
\n
Default Tables for IPv6 Search:
{chr(10).join([f'{idx + 1}. {_}' for idx, _ in enumerate(RANKED['ipv6'])])}
\n
Default Tables for MAC Search:
{chr(10).join([f'{idx + 1}. {_}' for idx, _ in enumerate(RANKED['mac'])])}
""",
    )
    arg_parser.add_argument(
        "address",
        help="IPv4, IPv6, or MAC address to search for.",
    )
    arg_parser.add_argument(
        "-f",
        "--full-scan",
        help="Scan all tables not just default.  It will take more time so it is "
        "recommended to use if first search could not locate anything.",
        action="store_true",
        default=False,
    )
    arg_parser.add_argument(
        "-r",
        "--regex",
        help="Perform a regex search instead of an exact match on the tables, please quote your regex.",
        choices=[
            "ipv4",
            "ipv6",
            "mac",
        ],
    )
    arg_parser.add_argument(
        "-B",
        "--break-on-match",
        help="Breaks out of searching tables when the first match is found.",
        action="store_true",
        default=False,
    )
    args = shared_args(arg_parser, "ipf_global_search")

    ipf = IPFClient(snapshot_id=args.snapshot, base_url=args.base_url, auth=args.auth)
    ipf._client.headers["user-agent"] += "; ipf_global_search"
    gs = GlobalSearch(client=ipf)
    if args.regex:
        results = gs.search_regex(
            search_type=args.regex, address=args.address, full_scan=args.full_scan, first_match=args.break_on_match
        )
    else:
        results = gs.search(address=args.address, full_scan=args.full_scan, first_match=args.break_on_match)
    print_results(list(results.values()), args.address, args)


def bold_string(results: list[dict], address: str):
    regex = re.compile(rf"({address})")

    def bold_loop(result):
        if isinstance(result, list):
            for _ in result:
                bold_loop(_)
        elif isinstance(result, dict):
            for k, v in result.items():
                result[k] = bold_loop(v)
        elif isinstance(result, (str, bytes)):
            return regex.sub(r"[b u]\1[/b u]", result)
        return result

    return bold_loop(deepcopy(results))


def rich_print(result: dict, address: str, count: bool = False):
    CONSOLE.print(f"\n{result['menu']}: {str(len(result['data']))}" if count else f"\n{result['menu']}")
    CONSOLE.print(result["url"], style=f"link {result['url']}")
    if not count:
        json_str = json.dumps(bold_string(result["data"], address), indent=4)
        CONSOLE.print(json_str)


def py_print(result: dict, count: bool = False):
    print(f"\n{result['menu']}: {str(len(result['data']))}" if count else f"\n{result['menu']}")
    print(result["url"])
    if not count:
        print(json.dumps(result["data"], indent=4))


def print_json(results: dict, address: str):
    if CONSOLE:
        for result in results:
            result["data"] = bold_string(result["data"], address)
        CONSOLE.print(json.dumps(results, indent=4))
    else:
        print(json.dumps(results, indent=4))


def print_results(results: list, address: str, args: argparse.Namespace):
    global CONSOLE
    if args.rich_disable:
        CONSOLE = None
    verbose, full_scan, count = args.verbose, args.full_scan, args.count
    if not verbose and not [y for x in results for y in x["data"]]:
        msg = f'\nAddress "{address}" not found in tables'
        msg += "." if full_scan else ", try running with '-f|--full-scan' to query all tables."
        CONSOLE.print(msg) if CONSOLE else LOGGER.error(msg)
        exit(1)
    if args.json:
        print_json(results, address)
        return
    for result in results:
        if not result["data"] and not verbose:
            continue
        if CONSOLE:
            rich_print(result, address, count)
        else:
            py_print(result, count)


if __name__ == "__main__":
    main()



from __future__ import annotations as _annotations

import os
import re
from ipaddress import IPv4Interface
from typing import Optional, Union, Any, Annotated

from pydantic import field_validator, BaseModel, Field, AliasChoices, model_serializer, StringConstraints
from pydantic.functional_validators import BeforeValidator, AfterValidator

from ipfabric.tools.shared import validate_ip_network_str, VALID_IP
from .constants import VALID_LAYOUTS

PYDANTIC_EXTRAS = os.getenv("IPFABRIC_PYDANTIC_EXTRAS", "allow")
PORT_REGEX = re.compile(r"^\d*$|^\d*-\d*$")
ALL_NETWORK = "$main"

IPv4 = Annotated[Union[str, VALID_IP], BeforeValidator(validate_ip_network_str)]


class Instance(BaseModel, extra=PYDANTIC_EXTRAS):
    rootId: str
    vlanId: int
    visible: bool = True
    grouped: bool = True


class STPInstances(BaseModel, extra=PYDANTIC_EXTRAS):
    isolate: bool = False
    instances: list[Instance] = Field(default_factory=list)


class Technologies(BaseModel, extra=PYDANTIC_EXTRAS):
    expandDeviceGroups: Optional[list[str]] = Field(default_factory=list)
    stpInstances: Optional[STPInstances] = Field(default_factory=STPInstances)


class ICMP(BaseModel, extra=PYDANTIC_EXTRAS):
    type: int
    code: int


class OtherOptions(BaseModel, extra=PYDANTIC_EXTRAS):
    applications: str = ".*"
    tracked: bool = False
    category: str = ""
    url: str = ""


class EntryPoint(BaseModel, extra=PYDANTIC_EXTRAS):
    sn: str = Field(title="Serial Number", description="IP Fabric Unique Device Serial Number API column sn")
    iface: str = Field(
        title="Interface", description="Interface to use as entry point. This is the intName not nameOriginal."
    )
    hostname: str = Field(title="Hostname", description="Hostname of the Device")


class Algorithm(BaseModel):
    """Default is automatic. Adding entryPoints will change to userDefined."""

    vrf: Optional[str] = None
    entryPoints: Optional[list[EntryPoint]] = None

    @property
    def type(self):
        return "userDefined" if self.entryPoints else "automatic"

    @model_serializer
    def _serialize_algorithm(self) -> dict[str, Any]:
        if self.entryPoints:
            return dict(type=self.type, entryPoints=[_.model_dump() for _ in self.entryPoints])
        else:
            return dict(type=self.type, vrf=self.vrf) if self.vrf else dict(type=self.type)


class PathLookup(BaseModel, extra=PYDANTIC_EXTRAS):
    protocol: Annotated[str, AfterValidator(lambda s: s.lower()), StringConstraints(pattern=r"tcp|udp|icmp")] = Field(
        "tcp", title="Protocol", description="Valid protocols are tcp, udp, or icmp."
    )
    srcPorts: Union[str, int] = Field(
        "1024-65535",
        title="Source Ports",
        description="Source ports if protocol is tcp or udp. "
        "Can be comma separated, a range using -, or any combination.",
    )
    dstPorts: Union[str, int] = Field(
        "80,443",
        title="Destination Ports",
        description="Destination ports if protocol is tcp or udp. "
        "Can be comma separated, a range using -, or any combination.",
    )
    tcpFlags: list[
        Annotated[str, AfterValidator(lambda s: s.lower()), StringConstraints(pattern=r"ack|fin|psh|rst|syn|urg")]
    ] = Field(
        default_factory=list,
        title="TCP Flags",
        description="Optional additional flags if protocol = TCP. "
        "Valid flags are ['ack', 'fin', 'psh', 'rst', 'syn', 'urg']",
        validation_alias=AliasChoices("tcpFlags", "flags"),
    )
    icmp: ICMP = Field(
        ICMP(type=0, code=0),
        title="ICMP Packet",
        description="Default is Echo Reply (type=0, code=0). You can pass in an ICMP model from ipfabric.diagrams.icmp "
        "or specify your own values like {'type': 1, 'code': 2}.",
    )
    ttl: int = Field(128, title="Time To Live", description="TTL value, default is 128.")
    fragmentOffset: int = Field(0, title="Fragment Offset", description="Fragment Offset value, default is 0.")
    securedPath: bool = True
    enableRegions: bool = False
    srcRegions: str = ".*"
    dstRegions: str = ".*"
    otherOptions: OtherOptions = Field(default_factory=OtherOptions)
    firstHopAlgorithm: Algorithm = Field(default_factory=Algorithm)

    @field_validator("srcPorts", "dstPorts")
    @classmethod
    def _check_ports(cls, v):
        ports = str(v).replace(" ", "").split(",")
        for p in ports:
            if not PORT_REGEX.match(p):
                raise ValueError(
                    f'Ports "{v}" is not in the valid syntax, examples: ["80", "80,443", "0-1024", "80,8000-8100,8443"]'
                )
            if "-" in p:
                pn = p.split("-")
                if int(pn[0]) >= int(pn[1]):
                    raise ValueError(f'Ports "{p}" is invalid. {pn[0]} must be smaller than {pn[1]}.')
        return str(",".join(ports))

    @property
    def l4_options(self) -> dict[str, Any]:
        if self.protocol == "icmp":
            return dict(type=self.icmp.type, code=self.icmp.code)
        elif self.protocol == "udp":
            return dict(srcPorts=self.srcPorts, dstPorts=self.dstPorts)
        else:
            return dict(srcPorts=self.srcPorts, dstPorts=self.dstPorts, flags=self.tcpFlags)

    @model_serializer
    def _serializer(self) -> dict[str, Any]:
        return dict(
            type="pathLookup",
            groupBy="siteName",
            protocol=self.protocol,
            ttl=self.ttl,
            fragmentOffset=self.fragmentOffset,
            securedPath=self.securedPath,
            enableRegions=self.enableRegions,
            srcRegions=self.srcRegions,
            dstRegions=self.dstRegions,
            l4Options=self.l4_options,
            otherOptions=vars(self.otherOptions),
            firstHopAlgorithm=self.firstHopAlgorithm.model_dump(),
        )


class Multicast(PathLookup, BaseModel, extra=PYDANTIC_EXTRAS):
    group: IPv4
    source: IPv4
    receiver: Optional[IPv4] = None

    @field_validator("group")
    @classmethod
    def _valid_group(cls, v):
        ip = IPv4Interface(v)
        if not ip.network.is_multicast:
            raise ValueError(f'Group IP "{v}" is not a valid Multicast Address.')
        return v

    @model_serializer
    def _serializer(self) -> dict[str, Any]:
        parameters = super()._serializer()
        if self.receiver:
            parameters["receiver"] = str(self.receiver)
        return dict(
            pathLookupType="multicast",
            group=str(self.group),
            source=str(self.source),
            **parameters,
        )


class Unicast(PathLookup, BaseModel, extra=PYDANTIC_EXTRAS):
    startingPoint: IPv4 = Field(title="Source IP Address or Subnet")
    destinationPoint: IPv4 = Field(title="Destination IP Address or Subnet")

    @model_serializer
    def _serializer(self) -> dict[str, Any]:
        parameters = super()._serializer()
        return dict(
            pathLookupType="unicast",
            networkMode=self._check_subnets(),
            startingPoint=self.startingPoint,
            destinationPoint=self.destinationPoint,
            **parameters,
        )

    def _check_subnets(self) -> bool:
        """Checks for valid IP Addresses or Subnet.

        :return: True if a subnet is found to set to networkMode, False if only hosts
        """
        masks = {IPv4Interface(ip).network.prefixlen for ip in [self.startingPoint, self.destinationPoint]}
        return True if masks != {32} else False

    @property
    def swap_src_dst(self) -> Unicast:
        return self.model_copy(
            update={
                "startingPoint": self.destinationPoint,
                "destinationPoint": self.startingPoint,
                "srcPorts": self.dstPorts,
                "dstPorts": self.srcPorts,
                "srcRegions": self.dstRegions,
                "dstRegions": self.srcRegions,
            }
        )


class Host2GW(BaseModel, extra=PYDANTIC_EXTRAS):
    startingPoint: IPv4
    vrf: Optional[str] = None

    @model_serializer
    def _serializer(self) -> dict[str, Any]:
        parameters = dict(
            pathLookupType="hostToDefaultGW",
            type="pathLookup",
            groupBy="siteName",
            startingPoint=self.startingPoint,
        )
        if self.vrf:
            parameters["vrf"] = self.vrf
        return parameters


class Layout(BaseModel, extra=PYDANTIC_EXTRAS):
    path: str
    layout: str

    @field_validator("layout")
    @classmethod
    def _valid_layout(cls, v):
        if v and v not in VALID_LAYOUTS:
            raise ValueError(f'Layout "{v}" is not in the valid layouts of {VALID_LAYOUTS}')
        return v


class Network(BaseModel, extra=PYDANTIC_EXTRAS):
    sites: Optional[Union[str, list[str]]] = Field(default_factory=list)
    all_network: bool = Field(True, description="Show all sites as clouds, UI option 'All Network'")
    layouts: Optional[list[Layout]] = None
    technologies: Optional[Technologies] = None

    @field_validator("sites")
    @classmethod
    def _format_paths(cls, v):
        if isinstance(v, str):
            return [v]
        return v

    @model_serializer
    def _serializer(self) -> dict[str, Any]:
        parameters = dict(type="topology", groupBy="siteName", paths=self.sites.copy())
        if self.all_network and ALL_NETWORK not in parameters["paths"]:
            parameters["paths"].append(ALL_NETWORK)
        if self.layouts:
            parameters["layouts"] = [_.model_dump() for _ in self.layouts]
        if self.technologies:
            parameters["technologies"] = self.technologies.model_dump()
        return parameters



import logging
from collections import defaultdict
from typing import Any, Union, Optional

from httpx import HTTPStatusError
from pydantic import BaseModel, PrivateAttr, Field

from ipfabric.tools.shared import raise_for_status
from .intent_check import Group, IntentCheck

logger = logging.getLogger("ipfabric")
COLOR_DICT = dict(nan=-1, green=0, blue=10, amber=20, red=30)


class Intent(BaseModel):
    client: Any = Field(exclude=True)
    _intent_checks: list[IntentCheck] = PrivateAttr(default_factory=list)
    _groups: list[Group] = PrivateAttr(default_factory=list)
    _snapshot_id: Optional[str] = None

    def model_post_init(self, __context: Any) -> None:
        self._snapshot_id = self.client.snapshot_id
        if self.snapshot_id:
            try:
                self._intent_checks: list[IntentCheck] = self.get_intent_checks()
            except ValueError:
                pass
            self._groups: list[Group] = self.get_groups()

    @property
    def intent_checks(self):
        if not self._intent_checks:
            self.load_intent(self.client.snapshot_id)
        return self._intent_checks

    @property
    def groups(self):
        if not self._groups:
            self.load_intent(self.client.snapshot_id)
        return self._groups

    @property
    def snapshot_id(self):
        return self._snapshot_id

    @snapshot_id.setter
    def snapshot_id(self, snapshot_id: str):
        self._snapshot_id = snapshot_id

    def get_intent_checks(self, snapshot_id: str = None) -> list[IntentCheck]:
        """Gets all intent checks and returns a list of them.

        Args:
        snapshot_id: Optional snapshot ID to get different data

        Returns:
            list: List of intent checks
        """
        snapshot = self.client.snapshots[snapshot_id] if snapshot_id else self.client.snapshots[self.snapshot_id]
        if not snapshot.loaded:
            raise ValueError(f"Snapshot {snapshot.snapshot_id} is not loaded; cannot pull Intent Rules.")
        if snapshot.disabled_intent_verification is True:
            raise ValueError(
                f"Snapshot {snapshot.snapshot_id} has Intent Verification computation disabled; "
                "cannot pull Intent Rules."
            )
        res = raise_for_status(self.client.get("reports", params=dict(snapshot=snapshot.snapshot_id)))
        try:
            return [IntentCheck(**check) for check in res.json()]
        except HTTPStatusError:
            logger.warning(self.client._api_insuf_rights + 'on GET "/reports". Will not load Intents.')
            return list()

    def load_intent(self, snapshot_id: str = None):
        """Loads intent checks into the class.

        Args:
            snapshot_id: Uses a different Snapshot ID then client
        """
        self.snapshot_id = snapshot_id or self.client.snapshot_id
        self._intent_checks = self.get_intent_checks(snapshot_id)
        self._groups = self.get_groups()

    def get_groups(self) -> list:
        """
        Returns:
            list: list of groups
        """
        res = self.client.get("reports/groups")
        if res.status_code == 200:
            return [Group(**group) for group in res.json()]
        else:  # TODO: Fix this error
            logger.warning(self.client._api_insuf_rights + 'on GET "/reports/groups". Will not load Intent Groups.')
            return list()

    @property
    def custom(self) -> list[IntentCheck]:
        return [c for c in self.intent_checks if c.custom]

    @property
    def builtin(self) -> list[IntentCheck]:
        return [c for c in self.intent_checks if not c.custom]

    @property
    def intent_by_id(self) -> dict[str, IntentCheck]:
        return {c.intent_id: c for c in self.intent_checks}

    @property
    def intents_by_name(self) -> dict[str, list[IntentCheck]]:
        reports = defaultdict(list)
        [reports[c.name].append(c) for c in self.intent_checks]
        return dict(reports)

    @property
    def intent_ids_by_web_endpoint(self) -> dict[str, list[str]]:
        reports = defaultdict(list)
        [reports[c.web_endpoint].append(c.intent_id) for c in self.intent_checks]
        return dict(reports)

    @property
    def intent_ids_by_api_endpoint(self) -> dict[str, list[str]]:
        reports = defaultdict(list)
        [reports[c.api_endpoint].append(c.intent_id) for c in self.intent_checks]
        return dict(reports)

    @property
    def group_by_id(self) -> dict[str, Group]:
        return {g.group_id: g for g in self.groups}

    @property
    def group_by_name(self) -> dict[str, Group]:
        return {g.name: g for g in self.groups}

    def get_results(self, intent: IntentCheck, color: Union[str, int], snapshot_id: str = None) -> list:
        """Get the outcome of an Intent Check by a specific color

        Args:
            intent: an IntentCheck, please see the Intent Check Model
            color: color of intent check
            snapshot_id: Uses a different Snapshot ID then client

        Returns:
            list: List of Dictionary objects.
        """
        if isinstance(color, str):
            color = COLOR_DICT[color]
        return self._get_data(intent, snapshot_id or self.snapshot_id, color)

    def get_all_results(self, intent: IntentCheck, snapshot_id: str = None):
        """set the intent check attributes

        Args:
            intent: an IntentCheck, please see the Intent Check Model
            snapshot_id: Uses a different Snapshot ID then client

        Returns:
            list: List of Dictionary objects.
        """
        snapshot_id = snapshot_id or self.snapshot_id
        for color_str, color_int in COLOR_DICT.items():
            if color_int == -1 or getattr(intent.result.checks, color_str):
                setattr(intent.result_data, color_str, self._get_data(intent, snapshot_id, color_int))
        return intent

    def _get_data(self, intent: IntentCheck, snapshot_id: str, color: int):
        return self.client.fetch_all(
            intent.api_endpoint,
            snapshot_id=snapshot_id,
            reports=intent.web_endpoint,
            filters={intent.column: ["color", "eq", color]},
        )

    def compare_snapshot(self, snapshot_id: str, reverse: bool = False) -> list:
        """Compares all intents against another snapshot.
        Current is the snapshot loaded into the class
        Other is the snapshot specified in this method.  Use reverse=True to flip them.

        Args:
            snapshot_id: Snapshot ID to compare against this will be the "other" key
            reverse: Default False, setting to true will flip current and other.
        Returns:
            list: List of dictionaries
        """
        new_intents = {i.name: i for i in self.get_intent_checks(snapshot_id)}
        comparison = list()
        for name, intent in new_intents.items():
            old = self.intents_by_name[name][0].result
            compare = intent.result.compare(old) if reverse else old.compare(intent.result)
            for desc, value in compare.items():
                n = desc if desc != "count" else "total"
                comparison.append({"name": name, "id": intent.intent_id, "check": n, **value})
        return comparison



import logging
from typing import Optional, Literal

from pydantic import Field, BaseModel, computed_field

from ipfabric.models.oas import Endpoint

logger = logging.getLogger("ipfabric")


class RBAC(BaseModel):
    name: str
    description: Optional[str] = None
    system: bool = Field(None, alias="isSystem")

    def __repr__(self):
        return f"{self.name}"


class Policy(RBAC, BaseModel):
    policy_id: str = Field(None, alias="id")
    scope_type: Literal["attributeScopes", "apiScopes"] = Field(None, alias="scopeType")
    api_scope_ids: Optional[list[str]] = Field(default_factory=list, alias="apiScopeIds")
    attribute_scope_ids: Optional[list[str]] = Field(default_factory=list, alias="attributeScopeIds")
    attribute_filters: Optional[dict] = Field(default_factory=dict, alias="attributeFilters")
    api_scopes: Optional[list[Endpoint]] = Field(default_factory=list)

    @computed_field
    @property
    def api_scopes_by_id(self) -> dict[str, Endpoint]:
        return {_.api_scope_id: _ for _ in self.api_scopes}

    def __hash__(self):
        return hash(self.policy_id)


class Role(RBAC, BaseModel):
    role_id: str = Field(None, alias="id")
    role_type: str = Field(None, alias="type")
    admin: bool = Field(None, alias="isAdmin")
    policy_ids: Optional[list[str]] = Field(default_factory=list, alias="policyIds")
    policies: Optional[list[Policy]] = Field(default_factory=list)

    @computed_field
    @property
    def policies_by_name(self) -> dict[str, Policy]:
        return {_.name: _ for _ in self.policies} if self.policies else dict()

    @computed_field
    @property
    def policies_by_id(self) -> dict[str, Policy]:
        return {_.policy_id: _ for _ in self.policies} if self.policies else dict()

    def __hash__(self):
        return hash(self.role_id)



from time import sleep
from typing import Optional, Union

from httpx import Client, ReadTimeout, HTTPStatusError
from httpx._types import ProxyTypes, TimeoutTypes
from pydantic import BaseModel, ConfigDict, field_validator

UNSUPPORTED_VENDORS = [
    "azure",
    "aws",
    "gcp",
    "brocade",
    "dell",
    "extreme",  # TODO: Verify
    "fs",
    "huawei",
    "nokia",
    "riverbed",
    "ruckus",
]
UNSUPPORTED_FAMILIES = [
    "meraki",
    "mist",
    "encs",
    "sg",
    "viptela",
    "lap",
    "prisma",
    "comware",
    "comware-lap",
    "3com",
    "velocloud",
]


class CveMetricV2(BaseModel):
    model_config = ConfigDict(extra="ignore")
    version: str
    accessVector: str
    accessComplexity: str
    confidentialityImpact: str
    integrityImpact: str
    availabilityImpact: str
    baseScore: float
    baseSeverity: str
    exploitabilityScore: float
    impactScore: float
    acInsufInfo: bool
    obtainAllPrivilege: bool
    obtainUserPrivilege: bool
    obtainOtherPrivilege: bool
    userInteractionRequired: Optional[bool] = None


class CveMetricV3(BaseModel):
    model_config = ConfigDict(extra="ignore")
    version: str
    attackVector: str
    attackComplexity: str
    privilegesRequired: str
    userInteraction: str
    scope: str
    confidentialityImpact: str
    integrityImpact: str
    availabilityImpact: str
    baseScore: float
    baseSeverity: str
    exploitabilityScore: float
    impactScore: float


class CPE(BaseModel):
    cpeName: str
    cpeNameId: str
    deprecated: bool
    created: str
    lastModified: str
    error: Optional[str] = None


class CVE(BaseModel):
    cve_id: str
    description: str
    url: str
    metric_v2: Optional[Union[CveMetricV2, dict]] = None
    metric_v3: Optional[Union[CveMetricV3, dict]] = None

    @field_validator("metric_v3")
    @classmethod
    def create_metric_v3(cls, metric: Union[dict, CveMetricV3]) -> Union[None, CveMetricV3]:
        if isinstance(metric, CveMetricV3):
            return metric
        elif not metric:
            return None
        metrics = sorted([_ for _ in metric.keys() if "cvssMetricV3" in _], reverse=True)
        if not metrics:
            return None
        return CveMetricV3(**metric[metrics[0]][0], **metric[metrics[0]][0]["cvssData"])

    @field_validator("metric_v2")
    @classmethod
    def create_metric_v2(cls, metric: Union[dict, CveMetricV2]) -> Union[None, CveMetricV2]:
        if isinstance(metric, CveMetricV2):
            return metric
        elif not metric:
            return None
        metrics = sorted([_ for _ in metric.keys() if "cvssMetricV2" in _], reverse=True)
        if not metrics:
            return None
        return CveMetricV2(**metric[metrics[0]][0], **metric[metrics[0]][0]["cvssData"])

    @property
    def base_score(self) -> Union[float, None]:
        if self.metric_v3:
            return self.metric_v3.baseScore
        elif self.metric_v2:
            return self.metric_v2.baseScore
        return None

    def __repr__(self):
        return self.cve_id

    def __hash__(self):
        return hash(self.cve_id)


class CVEs(BaseModel):
    cpe: Optional[CPE] = None
    total_results: int
    cves: list[CVE]
    error: Optional[str] = None


class NIST(Client):
    def __init__(
        self, nvd_api_key: str, timeout: TimeoutTypes = 60, proxies: Optional[ProxyTypes] = None, retries: int = 2
    ):
        """
        NIST updated to API v2.0.  You must request and pass an API Key which can be obtained at
        https://nvd.nist.gov/developers/request-an-api-key

        Args:
            nvd_api_key: str: https://nvd.nist.gov/developers/request-an-api-key
            timeout: Timeout value for requests
            proxies: Proxies for requests
            retries: int: Number of retries for requests, default 2
        """
        super().__init__(
            base_url="https://services.nvd.nist.gov/rest/json/",
            timeout=timeout,
            headers={"apiKey": nvd_api_key},
            proxy=proxies,
        )
        self.retries = retries

    def _check_cisco(self, family, version):
        safe_version = version.replace("(", "\\(").replace(")", "\\)")
        family_map = {
            "viptela": f"cisco:sd-wan:{safe_version}",
            # "asa": f"cisco:asa:{version.replace('(', '.').replace(')', '.')}",  # Not Supported requires Model
            "ftd": f"cisco:firepower_threat_defense:{version.split(' ')[0]}",
            "wlc-air": f"cisco:wireless_lan_controller_software:{version}",
            "apic": f"cisco:application_policy_infrastructure_controller:{safe_version}",
            "nx-os": f"cisco:nx-os:{safe_version}",
            "aci": f"cisco:nx-os:{safe_version}",
            "ios-xe": f"cisco:ios_xe:{self._normalize_ios_version(version, family, safe_version)}",
            "ios-xr": f"cisco:ios_xr:{safe_version}",
            "ios": f"cisco:ios:{self._normalize_ios_version(version, family, safe_version)}",
        }
        return family_map.get(family, False)

    @staticmethod
    def _normalize_ios_version(version, family, safe_version):
        if family == "ios-xe":
            parts = [part.lstrip("0") for part in version.split(".")]
            if "" in parts:
                parts.remove("")
            return ".".join(parts[:-1]) + parts[-1] if parts[-1].isalpha() else ".".join(parts)
        elif family == "ios" and ("CML" in version or len(version.split(":")) > 1):
            return version.split("(")[0]
        elif family == "ios" and version.count("(") == 2:
            _ = safe_version.find("\\(", safe_version.find(")"))
            return safe_version[:_] + version.split("(")[-1].rstrip(")")
        return safe_version

    @staticmethod
    def _check_hp(family, version):
        if family in ["arubacx", "arubasw"]:  # Verified
            _ = version.split(".", maxsplit=1)
            version = version if len(_) < 2 else _[1]
            family = "arubaos-cx" if family == "arubacx" else "arubaos-switch"
            return "hpe:" + family + ":" + version
        elif family == "aruba":  # Verified
            return "arubanetworks:arubaos:" + version
        elif family in ["aruba-iap", "aruba-lap"]:  # Verified
            return "arubanetworks:instant:" + version
        return False

    @staticmethod
    def _check_juniper(family, version):
        if "R" in version:
            _ = version.split("R")
            version = _[0] + ":r" + _[1].split(".")[0]
        elif "-D" in version:
            _ = version.split("-D")
            version = _[0] + ":d" + _[1].split(".")[0]
        return "juniper:" + family + ":" + version

    @staticmethod
    def _trim_version(version: str, max_len: int = 3) -> str:
        parts = version.split(".")
        return ".".join(parts[:max_len]) if len(parts) > max_len else version

    def get_cpe(self, vendor: str, family: str, version: str) -> list[CPE]:  # noqa: C901
        """returns CVE data about a specific product line

        Args:
        vendor: Vendor for the device to be checked
        family: Family of the device to be checked
        version: Software version of the device to be checked
        Returns:
            a list of CVEs
        """
        match_str = "cpe:2.3:*:"
        unsupported = [
            CPE(cpeName="", cpeNameId="", deprecated=False, created="", lastModified="", error="Unsupported")
        ]
        if vendor in UNSUPPORTED_VENDORS or family in UNSUPPORTED_FAMILIES:
            return unsupported

        vendor_family_map = {
            "pan-os": f"paloaltonetworks:pan-os:{':'.join(version.split('-'))}",
            "alcatel": f"alcatel:{family}:{version.split('.R')[0]}",
            "adc": f"citrix:application_delivery_controller:{version}",
            "big-ip": f"f5:big-ip_local_traffic_manager:{version}",
            "fortigate": f"fortinet:fortios:{version.split(',')[0]}",
            "fortiswitch": f"fortinet:fortiswitch:{version.split(',')[0]}",
            "forcepoint": f"forcepoint:next_generation_firewall:{self._trim_version(version, 3)}",
            "frr": f"frrouting:frrouting:{version}",
            "checkpoint": f"checkpoint:gaia_os:{version}",
            "mikrotik": f"mikrotik:{family}:{version}",
            "opengear": f"opengear:opengear:{version}",
            "quagga": f"quagga:quagga:{version}",
            "arista": f"arista:{family}:{version.split('-')[0]}",
            "versa": f"versa-networks:versa_operating_system:{version.split('-')[0]}",
            "silverpeak": f"arubanetworks:edgeconnect_enterprise:{version.split('_')[0]}",
            "stormshield": f"stormshield:stormshield_network_security:{version}",
            "nsx-t": f"vmware:nsx-t_data_center:{self._trim_version(version, 5)}",
        }

        if vendor in ["cisco", "hp"]:
            _ = self._check_cisco(family, version) if vendor == "cisco" else self._check_hp(family, version)
            if not _:
                return unsupported
            match_str += _
        elif vendor == "juniper":
            match_str += self._check_juniper(family, version)
        elif vendor in vendor_family_map:
            match_str += vendor_family_map[vendor]
        elif family in vendor_family_map:
            match_str += vendor_family_map[family]
        else:
            return unsupported
        return self._query_cpe({"cpeMatchString": match_str + ":", "startIndex": 0})

    def check_cve(self, vendor: str, family: str, version: str) -> list[CVEs]:
        cpes = self.get_cpe(vendor, family, version)
        if not cpes:
            return [CVEs(total_results=0, cves=[], error="No CPEs Found.")]
        cves = list()
        for cpe in cpes:
            if cpe.error:
                cves.append(CVEs(total_results=0, cves=[], error=cpe.error, cpe=cpe))
            else:
                cves.append(self._query_cve(cpe))
        return cves

    def _query_cve(self, cpe: CPE, attempt: int = 0) -> CVEs:
        try:
            sleep(0.3 + attempt)  # NIST Rate Limit: 50 requests/30 seconds
            res = self.get("cves/2.0", params={"cpeName": cpe.cpeName, "startIndex": 0})
            res.raise_for_status()
            data = res.json()

            cves = CVEs(
                total_results=data["totalResults"],
                cves=[
                    CVE(
                        cve_id=i["cve"]["id"],
                        description=i["cve"]["descriptions"][0]["value"],
                        url=i["cve"]["references"][0]["url"],
                        metric_v2=i["cve"].get("metrics", {}),
                        metric_v3=i["cve"].get("metrics", {}),
                    )
                    for i in data["vulnerabilities"]
                ],
                cpe=cpe,
            )
            return cves
        except ReadTimeout:
            return CVEs(total_results=0, cves=[], error="Timeout", cpe=cpe)
        except HTTPStatusError:
            if attempt > self.retries:
                return CVEs(total_results=0, cves=[], error="HTTP Error", cpe=cpe)
            else:
                return self._query_cve(cpe, attempt + 1)

    def _query_cpe(self, params, attempt: int = 0) -> list[CPE]:
        try:
            sleep(0.3 + attempt)  # NIST Rate Limit: 50 requests/30 seconds
            res = self.get("cpes/2.0", params=params)
            res.raise_for_status()
            data = res.json()
            return [CPE(**_["cpe"]) for _ in data["products"] if "cpe" in _]
        except ReadTimeout:
            return [CPE(cpeName="", cpeNameId="", deprecated=False, created="", lastModified="", error="Timeout")]
        except HTTPStatusError:
            if attempt > self.retries:
                return [
                    CPE(cpeName="", cpeNameId="", deprecated=False, created="", lastModified="", error="HTTP Error")
                ]
            else:
                return self._query_cpe(params, attempt + 1)



import logging
import os
from argparse import ArgumentParser, Namespace
from pathlib import Path

import dotenv

from ipfabric.tools.shared import valid_snapshot


def load_env():
    dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True) or Path("~").expanduser().joinpath(".env"))


def check_snapshot_arg(snapshot: str) -> str:
    if snapshot.lower() in ["last", "prev", "lastlocked"]:
        snapshot = "$" + snapshot
    snapshot = valid_snapshot(snapshot).strip("'")
    return snapshot


def get_auth_from_args_or_env(args):
    """
    Determine the authentication method based on args or environment variables.
    """
    if args.token:
        return args.token
    if args.username and args.password:
        return args.username, args.password
    if os.getenv("IPF_TOKEN"):
        return os.getenv("IPF_TOKEN")
    if os.getenv("IPF_USERNAME") and os.getenv("IPF_PASSWORD"):
        return os.getenv("IPF_USERNAME"), os.getenv("IPF_PASSWORD")
    else:
        raise ValueError("Authentication details must be provided as an argument or set as an environment variable.")


def base_args(arg_parser: ArgumentParser) -> ArgumentParser:
    arg_parser.add_argument(
        "-s",
        "--snapshot",
        help="Snapshot to use which can be a UUID or one of ['last', 'prev', 'lastLocked']"
        "with or without `$` for *nix compatability.",
        default="$last",
    )
    arg_parser.add_argument(
        "-b",
        "--base-url",
        default=os.getenv("IPF_URL", None),
        help="Base URL for IPFabric API or IPF_URL env variable.",
    )
    arg_parser.add_argument("-t", "--token", help="IP Fabric API Token or IPF_TOKEN env variable.")
    arg_parser.add_argument(
        "-u",
        "--username",
        help="IP Fabric username (requires --password) or IPF_USERNAME env variable.",
    )
    arg_parser.add_argument("-p", "--password", help="IP Fabric password or IPF_PASSWORD env variable.")
    return arg_parser


def parse_base_args(args: Namespace) -> Namespace:
    if not args.base_url:
        raise ValueError("Base URL must be provided as an argument or set as an environment variable (IPF_URL).")
    args.snapshot = check_snapshot_arg(args.snapshot)
    setattr(args, "auth", get_auth_from_args_or_env(args))
    return args


def shared_args(arg_parser: ArgumentParser, logger_name: str) -> Namespace:
    arg_parser = base_args(arg_parser)
    arg_parser.add_argument(
        "-c",
        "--count",
        help="Print count of rows instead of the actual data.",
        action="store_true",
        default=False,
    )
    arg_parser.add_argument(
        "-v",
        "--verbose",
        help="Verbose output will enable debugging and print all tables even if no data.",
        action="store_true",
        default=False,
    )
    arg_parser.add_argument(
        "-j",
        "--json",
        help="Enable JSON output which will also print all tables even if no data.",
        action="store_true",
        default=False,
    )
    arg_parser.add_argument(
        "-R",
        "--rich-disable",
        help="Disable rich formatting if installed. Useful for sending JSON output to jq.",
        action="store_true",
        default=False,
    )
    args = arg_parser.parse_args()

    logger = logging.getLogger(logger_name)
    if args.verbose:
        logger.setLevel(logging.DEBUG)
        logging.getLogger("ipfabric").setLevel(logging.DEBUG)
        logger.debug("Logging level set to DEBUG")

    args = parse_base_args(args)
    logger.debug(f"Snapshot ID selected: {args.snapshot}")
    return args



import logging
import os
import tempfile
from typing import Optional, Literal, Any, Union
from urllib.parse import urlparse

import httpx
from httpx import Response, HTTPError
from pydantic import BaseModel, Field

from ipfabric.tools.shared import raise_for_status
from ipfabric.models.users import User

logger = logging.getLogger("ipfabric")


class Resources(BaseModel):
    cpu: int
    memory: int


class Extension(BaseModel):
    client: Optional[Any] = Field(None, exclude=True)
    id: str
    name: str
    description: Optional[str] = None
    slug: str
    status: str
    type: Optional[Literal["docker-zip", "docker-image"]] = None
    environment_variables: Optional[list[dict[str, str]]] = Field(alias="environmentVariables")
    access_permission: str = Field(alias="accessPermission")
    accessible_by_users: list[str] = Field(alias="accessibleByUsers")
    resources: Resources

    @property
    def users(self) -> list[User]:
        """
        Fetch users with access to the extension.

        Returns:
            List[User]: List of users with access to the extension
        """
        try:
            return [self.client.settings.local_users.users_by_id[user] for user in self.accessible_by_users]
        except:  # noqa: E722
            return []


class Extensions(BaseModel):
    client: Any = Field(exclude=True)

    @property
    def extensions(self) -> list[Extension]:
        """
        Fetch all extensions.

        Returns:
            List[Extension]: List of all available extensions
        """
        response = raise_for_status(self.client.get("extensions")).json()["extensions"]
        return [Extension(client=self.client, **ext) for ext in response]

    def extension_by_id(self, extension_id: str) -> Extension:
        """
        Fetch a specific extension by ID.

        Args:
            extension_id (str): The ID of the extension to fetch

        Returns:
            Extension: The requested extension
        """
        return Extension(client=self.client, **raise_for_status(self.client.get(f"extensions/{extension_id}")).json())

    def logs_by_extension_id(self, extension_id: str) -> dict:
        """
        Fetch paginated logs for a specific extension.

        Args:
            extension_id (str): The ID of the extension to fetch logs for

        Returns:
            Dict: The paginated logs response
        """
        return raise_for_status(self.client.get(f"extensions/{extension_id}/logs")).json()

    def extension_by_name(self, name: str) -> Extension:
        """
        Fetch a specific extension by name.
        """
        extension = next((ext for ext in self.extensions if ext.name == name), None)
        if extension is None:
            raise ValueError(f"Extension with name '{name}' not found")
        return extension

    def extension_by_slug(self, slug: str) -> Extension:
        """
        Fetch a specific extension by slug.
        """
        extension = next((ext for ext in self.extensions if ext.slug == slug), None)
        if extension is None:
            raise ValueError(f"Extension with slug '{slug}' not found")
        return extension

    @staticmethod
    def _users_permissions(
        access_permission: Literal["anyone", "only-me", "selected-users"] = "anyone",
        accessible_by_users: Optional[list[str]] = None,
    ):
        if access_permission != "selected-users":
            if accessible_by_users:
                logger.warning("Accessible by Users is only allowed with selected-users permission")
            return {"accessPermission": access_permission, "accessibleByUsers": []}
        return {"accessPermission": access_permission, "accessibleByUsers": accessible_by_users}

    def _register_extension(
        self,
        docker_type: Literal["docker-zip", "docker-image"],
        file: bytes,
        name: str,
        slug: str,
        description: str,
        cpu: int = 1,
        memory: int = 512,
        environment_variables: Optional[dict[str, str]] = None,
        access_permission: Literal["anyone", "only-me", "selected-users"] = "anyone",
        accessible_by_users: Optional[list[str]] = None,
    ) -> Response:
        files = (
            {"file": ("extension.zip", file, "application/zip")}
            if docker_type == "docker-zip"
            else {"file": ("image.tar", file, "application/x-tar")}
        )

        data = {
            "name": name,
            "slug": slug,
            "description": description,
            "resources": {"cpu": cpu, "memory": memory},
            **self._users_permissions(access_permission, accessible_by_users),
        }

        if environment_variables:
            data["environmentVariables"] = [
                {"name": key, "value": value} for key, value in environment_variables.items()
            ]

        return raise_for_status(self.client.post(f"extensions/{docker_type}", files=files, data=data, timeout=300))

    def register_docker_zip(
        self,
        file: bytes,
        name: str,
        slug: str,
        description: str,
        cpu: int = 1,
        memory: int = 512,
        environment_variables: Optional[dict[str, str]] = None,
        access_permission: Literal["anyone", "only-me", "selected-users"] = "anyone",
        accessible_by_users: Optional[list[str]] = None,
    ) -> Response:
        """
        Register a new extension using zipped source code. Raises an exception if the extension is not registered successfully.

        Args:
            file (bytes): The zipped source code file
            name (str): Name of the extension
            slug (str): Slug for the extension
            description (str): Description of the extension
            cpu (int): Number of CPU cores for the extension
            memory (int): Memory size in MB for the extension
            environment_variables (Optional[Dict[str, str]]): Environment variables for the extension
            access_permission (Literal["anyone", "only-me", "selected-users"]): Access permission for the extension
            accessible_by_users (Optional[List[str]]): List of User IDs with access to the extension
        """
        return self._register_extension(
            "docker-zip",
            file=file,
            name=name,
            slug=slug,
            description=description,
            cpu=cpu,
            memory=memory,
            environment_variables=environment_variables,
            access_permission=access_permission,
            accessible_by_users=accessible_by_users,
        )

    def register_docker_image(
        self,
        file: bytes,
        name: str,
        slug: str,
        description: str,
        cpu: int = 1,
        memory: int = 512,
        environment_variables: Optional[dict[str, str]] = None,
        access_permission: Literal["anyone", "only-me", "selected-users"] = "anyone",
        accessible_by_users: Optional[list[str]] = None,
    ) -> Response:
        """
        Register a new extension using a docker image. Raises an exception if the extension is not registered successfully.

        Args:
            file (bytes): The tar file of the docker image
            name (str): Name of the extension
            slug (str): Slug for the extension
            description (str): Description of the extension
            cpu (int): Number of CPU cores for the extension
            memory (int): Memory size in MB for the extension
            environment_variables (Optional[Dict[str, str]]): Environment variables for the extension
            access_permission (Literal["anyone", "only-me", "selected-users"]): Access permission for the extension
            accessible_by_users (Optional[List[str]]): List of User IDs with access to the extension
        """
        self._register_extension(
            "docker-zip",
            file=file,
            name=name,
            slug=slug,
            description=description,
            cpu=cpu,
            memory=memory,
            environment_variables=environment_variables,
            access_permission=access_permission,
            accessible_by_users=accessible_by_users,
        )

    def start_extension(self, extension_id: str) -> Response:
        """
        Start an extension by its ID. Raises an exception if the extension fails to start.

        Args:
            extension_id (str): The ID of the extension to start
        """
        return raise_for_status(self.client.post(f"extensions/{extension_id}/start"))

    def stop_extension(self, extension_id: str) -> Response:
        """
        Stop an extension by its ID. Raises an exception if the extension fails to stop.

        Args:
            extension_id (str): The ID of the extension to stop
        """
        return raise_for_status(self.client.post(f"extensions/{extension_id}/stop"))

    def unregister_extension(self, extension_id: str) -> Response:
        """
        Unregister an extension by its ID. Raises an exception if the extension fails to unregister.

        Args:
            extension_id (str): The ID of the extension to unregister
        """
        return raise_for_status(self.client.delete(f"extensions/{extension_id}"))

    def register_from_git_url(
        self,
        git_url: str,
        name: str,
        slug: str,
        description: str,
        branch: str = "main",
        cpu: int = 1,
        memory: int = 512,
        environment_variables: Optional[dict[str, str]] = None,
        access_permission: Literal["anyone", "only-me", "selected-users"] = "anyone",
        accessible_by_users: Optional[list[str]] = None,
    ) -> Union[Response, None]:
        """
        Register an extension from a Git repository URL. Supports GitHub and GitLab repositories.
        Downloads the repository, creates a ZIP file, and registers it as a docker-zip extension.

        Args:
            git_url (str): URL to the Git repository (GitHub or GitLab)
            name (str): Name of the extension
            slug (str): Slug for the extension
            description (str): Description of the extension
            branch (str): Branch to download (default: "main")
            cpu (int): Number of CPU cores for the extension
            memory (int): Memory size in MB for the extension
            environment_variables (Optional[Dict[str, str]]): Environment variables for the extension
            access_permission (Literal["anyone", "only-me", "selected-users"]): Access permission for the extension
            accessible_by_users (Optional[List[str]]): List of User IDs with access to the extension

        Example:
            client.extensions.register_from_git_url(
                "https://github.com/username/repo",
                "My Extension",
                "my-extension",
                "Description",
                branch="main"
            )
        """
        parsed_url = urlparse(git_url)

        path_parts = parsed_url.path.rstrip(".git").strip("/").split("/")

        if "github.com" in parsed_url.netloc:
            if len(path_parts) != 2:
                raise ValueError("Invalid GitHub URL format")
            user, repo = path_parts
            download_url = f"https://github.com/{user}/{repo}/archive/{branch}.zip"

        elif "gitlab.com" in parsed_url.netloc:
            if len(path_parts) < 2:
                raise ValueError("Invalid GitLab URL format")
            project = path_parts[-1]  # Last part is the project name
            group = "/".join(path_parts[:-1])  # Everything else is the group path
            download_url = f"https://gitlab.com/{group}/{project}/-/archive/{branch}/{project}-{branch}.zip"

        else:
            raise ValueError("Only GitHub and GitLab repositories are supported")

        response = httpx.get(download_url)
        if response.status_code != 200:
            raise HTTPError(f"Failed to download repository: {response.status_code}")

        with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
            tmp_file.write(response.content)

        try:
            with open(tmp_file.name, "rb") as f:
                resp = self._register_extension(
                    "docker-zip",
                    file=f.read(),
                    name=name,
                    slug=slug,
                    description=description,
                    cpu=cpu,
                    memory=memory,
                    environment_variables=environment_variables,
                    access_permission=access_permission,
                    accessible_by_users=accessible_by_users,
                )
        finally:
            os.unlink(tmp_file.name)
        return resp



import logging
from collections import OrderedDict
from datetime import datetime, timezone
from typing import Any, Union
from typing import OrderedDict as OrderedDictType
from uuid import UUID

from pydantic import BaseModel, PrivateAttr, Field

from ipfabric.models import Snapshot, SNAPSHOT_COLUMNS
from ipfabric.tools.shared import VALID_REFS, raise_for_status

logger = logging.getLogger("ipfabric")

LAST_ID, PREV_ID, LASTLOCKED_ID = VALID_REFS


class Snapshots(BaseModel):
    client: Any = Field(exclude=True)
    _snapshots: OrderedDictType[str, Snapshot] = PrivateAttr(default_factory=dict)

    def model_post_init(self, __context) -> None:
        self.snapshots = self.get_snapshots()
        self.client._last_snapshot_update = datetime.now(timezone.utc)

    def update(self):
        """get all snapshots and assigns them to an attribute"""
        self.snapshots = self.get_snapshots()
        self.client._last_snapshot_update = datetime.now(timezone.utc)
        self.client._no_loaded_snapshots = self.loaded_snapshots == dict()

    @property
    def snapshots(self) -> OrderedDictType[str, Snapshot]:
        return self._snapshots

    @snapshots.setter
    def snapshots(self, _):
        self._snapshots = _

    @property
    def loaded_snapshots(self) -> OrderedDictType[str, Snapshot]:
        """get only loaded snapshots"""
        return OrderedDict([(k, v) for k, v in self.snapshots.items() if v.loaded])

    @property
    def loading_snapshot(self) -> Union[Snapshot, None]:
        """Return Loading Snapshot or None"""
        return next(iter([v for v in self.snapshots.values() if v.loading]), None)

    @property
    def running_snapshot(self) -> Union[Snapshot, None]:
        """Return Running Snapshot"""
        return next(iter([v for v in self.snapshots.values() if v.running]), None)

    @property
    def unloaded_snapshots(self) -> OrderedDictType[str, Snapshot]:
        if not self.client.unloaded:
            logger.warning("Unloaded snapshots not initialized. Retrieving unloaded snapshots.")
            self.client.unloaded = True
            self.update()
        return OrderedDict([(k, v) for k, v in self.snapshots.items() if not v.loaded])

    def get_snapshot(self, snapshot_id: str) -> Snapshot:
        if snapshot_id in self.snapshots:
            return self.snapshots[snapshot_id]
        else:
            payload = {"columns": list(SNAPSHOT_COLUMNS), "filters": {"id": ["eq", snapshot_id]}}
            results = list(self.client._ipf_pager("tables/management/snapshots", payload))
            if not results:
                logger.error(f"Snapshot {snapshot_id} not found.")
                return None
            get_results = self._get_snapshots()
            snapshot = self._create_snapshot_model(results[0], get_results)
            if snapshot.loaded:
                snapshot.get_assurance_engine_settings()
            return snapshot

    def _create_snapshot_model(self, s, get_results):
        return Snapshot(
            client=self.client,
            **s,
            licensedDevCount=get_results[s["id"]].get("licensedDevCount", None),
            errors=get_results[s["id"]].get("errors", None),
            version=get_results[s["id"]]["version"],
            initialVersion=get_results[s["id"]].get("initialVersion", None),
        )

    def get_snapshot_id(self, snapshot: Union[Snapshot, str]) -> str:
        """
        Returns a Snapshot ID for a given input.

        Args:
            snapshot: Snapshot model, name, or ID

        Returns:
            Snapshot ID
        """
        if isinstance(snapshot, Snapshot):
            return snapshot.snapshot_id
        elif snapshot in VALID_REFS:
            return self.snapshots[snapshot].snapshot_id
        try:
            UUID(snapshot, version=4)
            return self.snapshots[snapshot].snapshot_id
        except ValueError:
            for snap in list(self.snapshots.values()):
                if snapshot == snap.name:
                    return snap.snapshot_id
        raise ValueError(f"Could not locate Snapshot ID for {snapshot}.")

    def _get_snapshots(self):
        """
        Need to do a GET and POST to get all Snapshot data. See NIM-7223
        POST Missing:
        licensedDevCount
        errors
        version
        initialVersion
        """
        res = raise_for_status(self.client.get("/snapshots"))
        return {s["id"]: s for s in res.json()}

    def get_snapshots(self) -> OrderedDictType[str, Snapshot]:
        """Gets all snapshots from IP Fabric and returns a dictionary of {ID: Snapshot_info}

        Returns:
            Dictionary with ID as key and dictionary with info as the value
        """
        payload = {"columns": list(SNAPSHOT_COLUMNS), "sort": {"order": "desc", "column": "tsEnd"}}
        if not self.client.unloaded:
            payload["filters"] = {"status": ["nreg", "unloaded|error"]}
        results = list(self.client._ipf_pager("tables/management/snapshots", payload))
        get_results = self._get_snapshots()

        snap_dict = OrderedDict()
        for s in results:
            snap = self._create_snapshot_model(s, get_results)
            snap_dict[snap.snapshot_id] = snap
            if snap.loaded:
                snap.get_assurance_engine_settings()
                if LASTLOCKED_ID not in snap_dict and snap.locked:
                    snap_dict[LASTLOCKED_ID] = snap
                if LAST_ID not in snap_dict:
                    snap_dict[LAST_ID] = snap
                    continue
                if PREV_ID not in snap_dict:
                    snap_dict[PREV_ID] = snap
        return snap_dict



import logging
from collections import defaultdict
from datetime import datetime
from ipaddress import IPv4Address, AddressValueError
from typing import Any, Union, Optional

from pydantic import BaseModel, Field

from ipfabric.tools.shared import date_parser, raise_for_status

logger = logging.getLogger("ipfabric")
DEFAULT_SNAPSHOT = "$last"
NAMED_SNAPSHOTS = [DEFAULT_SNAPSHOT, "$prev", "$first"]


def trigger_backup(ipf, sn: str = None, ip: str = None):
    if sn:
        payload = {"sn": sn}
    else:
        try:
            ip = str(IPv4Address(ip))
        except AddressValueError:
            raise ValueError(f"Invalid IP Address, CIDR is not allowed: {ip}")
        payload = {"ip": ip}
    return raise_for_status(ipf.post("/discovery/trigger-config-backup", json=payload)).status_code


class Config(BaseModel):
    config_id: str = Field(None, alias="id")
    sn: str
    hostname: str
    config_hash: str = Field(None, alias="hash")
    status: str
    last_change: datetime = Field(None, alias="lastChangeAt")
    last_check: datetime = Field(None, alias="lastCheckAt")
    text: Optional[str] = None

    def get_text_config(self, client: Any, sanitized: bool = True) -> str:
        """gets a device config
        Args:
            client: IPF Client
            sanitized: bool to determine to return sensitive data

        Returns:
            string containing config of a device
        """
        res = raise_for_status(
            client.get(
                "/tables/management/configuration/download",
                params=dict(hash=self.config_hash, sanitized=sanitized),
            )
        )
        self.text = res.text
        return res.text


class DeviceConfigs(BaseModel):
    client: Any = Field(exclude=True)

    def get_all_configurations(self, device: Optional[str] = None, sn: Optional[str] = None) -> Union[dict, None]:
        """Get all configurations in IP Fabric

        Args:
            device: Hostname (case insensitive) filter
            sn: Serial number of device
        Returns:
            results: dict: {sn: [Config, Config]}
        """
        if device or sn:
            filters = dict(sn=["eq", sn]) if sn else dict(hostname=["ieq", device])
            res = self.client.fetch_all(
                "tables/management/configuration",
                sort={"order": "desc", "column": "lastChangeAt"},
                columns=[
                    "id",
                    "sn",
                    "hostname",
                    "lastChangeAt",
                    "lastCheckAt",
                    "status",
                    "hash",
                ],
                filters=filters,
                snapshot=False,
            )
            if len(res) == 0:
                logger.warning(f"Could not find any configurations for device '{device}'.")
                return None
        else:
            res = self.client.fetch_all(
                "tables/management/configuration",
                sort={"order": "desc", "column": "lastChangeAt"},
                columns=[
                    "id",
                    "sn",
                    "hostname",
                    "lastChangeAt",
                    "lastCheckAt",
                    "status",
                    "hash",
                ],
                snapshot=False,
            )
        results = defaultdict(list)
        [results[cfg["sn"]].append(Config(**cfg)) for cfg in res]
        return results

    def _search_ip(self, ip: str, snapshot_id: str = None, log: bool = False) -> dict:
        res = self.client.fetch_all(
            "tables/addressing/managed-devs",
            columns=["ip", "hostname", "sn"],
            reports="/technology/addressing/managed-ip",
            filters=dict(ip=["eq", ip]),
        )
        if len(res) == 1 and not log:
            return {"hostname": res[0]["hostname"], "sn": res[0]["sn"]}
        if len(res) == 1 and log:
            res = self.client.inventory.devices.all(
                columns=["hostname", "taskKey", "sn"],
                snapshot_id=snapshot_id,
                filters=dict(sn=["eq", res[0]["sn"]]),
            )
            return {"hostname": res[0]["hostname"], "taskKey": res[0]["taskKey"], "sn": res[0]["sn"]}
        elif len(res) > 1:
            logger.warning(f"Found multiple entries for IP '{ip}'.")
        elif len(res) == 0:
            logger.warning(f"Could not find a matching IP for '{ip}'.")
        return {"hostname": None, "sn": None}

    def get_configuration(
        self, device: str = None, sn: str = None, sanitized: bool = True, date: Union[str, tuple] = DEFAULT_SNAPSHOT
    ) -> Union[Config, None]:
        """Gets last configuration of a device based on hostname or IP or IPF Unique Serial Number

        Args:
        device: Hostname or IP
        sn: Serial Number
        sanitized: Default True to mask passwords
        date: Defaults to latest config. Values in [$last, $prev, $first] or can be a
            tuple of a date range to get the latest snapshot in that range.
            Date can be string or int in seconds ("11/22/ 1:30", 1637629200)
        Returns:
            Returns a result or None
        """
        if not isinstance(date, tuple) and date not in NAMED_SNAPSHOTS:
            raise SyntaxError(f"Date must be in {NAMED_SNAPSHOTS} or tuple ('startDate', 'endDate')")
        if not sn:
            sn = self._validate_device(device)["sn"]
            if not sn:
                return None
        cfgs = self.get_all_configurations(sn=sn)
        if not cfgs:
            return None
        cfg = self._get_hash(cfgs[sn], date)
        if cfg:
            return self.get_text_config(cfg, sanitized)
        else:
            logger.error(f"Could not find a configuration with date {date}")
            return None

    def get_text_config(self, cfg: Config, sanitized: bool = True) -> Config:
        """gets a devices config
        Args:
            cfg: Config from get_configuration method
            sanitized: bool to determine to return sensitive data

        Returns:
            string containing config of a device
        """
        res = raise_for_status(
            self.client.get(
                "/tables/management/configuration/download",
                params=dict(hash=cfg.config_hash, sanitized=sanitized),
            )
        )
        cfg.text = res.text
        return cfg

    @staticmethod
    def _get_hash(configs, date):
        if isinstance(date, tuple):
            start = date_parser(date[0])
            end = date_parser(date[1])
            for cfg in configs:
                if start < cfg.last_change < end:
                    return cfg
        elif date == DEFAULT_SNAPSHOT:
            return configs[0]
        elif date == "$prev" and len(configs) > 1:
            return configs[1]
        elif date == "$first":
            return configs[-1]
        return None

    def _validate_device(self, device: str, snapshot_id: str = None, log: bool = False) -> dict:
        try:
            ip = IPv4Address(device)
            return self._search_ip(str(ip), snapshot_id=snapshot_id, log=log)
        except AddressValueError:
            res = self.client.inventory.devices.all(
                columns=["hostname", "taskKey", "sn"],
                filters=dict(hostname=["ieq", device]),
                snapshot_id=snapshot_id,
            )
            if len(res) == 1:
                return {"hostname": res[0]["hostname"], "taskKey": res[0]["taskKey"], "sn": res[0]["sn"]}
            elif len(res) == 0:
                logger.warning(f"Could not find a matching device for '{device}'")
            elif len(res) > 1:
                logger.warning(f"Found multiple devices matching '{device}'.")
        return {"hostname": None, "sn": None}

    def get_log(self, device: str, snapshot_id: str = None):
        device = self._validate_device(device, snapshot_id=snapshot_id, log=True)
        if not device["sn"]:
            return None
        return self.get_text_log(device)

    def get_text_log(self, device: dict):
        return raise_for_status(self.client.get("/os/logs/task/" + device["taskKey"])).text



from .configuration import DeviceConfigs, trigger_backup
from .discovery_history import DiscoveryHistory
from .feature_matrix import FeatureMatrix
from .restore_intents import RestoreIntents
from .shared import (
    parse_mac,
    TIMEZONES,
    VALID_REFS,
    validate_ip_network_str,
    VALID_IP,
    valid_slug,
    raise_for_status,
    valid_snapshot,
)
from .site_seperation_report import map_devices_to_rules
from .vulnerabilities import Vulnerabilities

__all__ = [
    "DeviceConfigs",
    "trigger_backup",
    "Vulnerabilities",
    "RestoreIntents",
    "DiscoveryHistory",
    "map_devices_to_rules",
    "parse_mac",
    "TIMEZONES",
    "VALID_REFS",
    "validate_ip_network_str",
    "VALID_IP",
    "valid_slug",
    "raise_for_status",
    "valid_snapshot",
    "FeatureMatrix",
]



import logging
from copy import deepcopy
from typing import Optional

from pydantic.dataclasses import dataclass

from ipfabric.settings.attributes import Attributes
from ipfabric.settings.site_separation import SiteSeparation

logger = logging.getLogger("ipfabric")


@dataclass
class Matches:
    hostname: str
    sn: str
    old_site_name: str
    rule_type: str
    slug: Optional[str] = None
    cloud_resource_id: Optional[str] = None
    new_site_name: Optional[str] = None
    rule_number: Optional[int] = None
    rule_note: Optional[str] = None
    regex: Optional[str] = None
    transformation: Optional[str] = None


def check_attributes(ipf, devices):
    matches = list()
    attributes = {a["sn"]: a for a in Attributes(client=ipf).all(filters={"name": ["eq", "siteName"]})}
    for sn, dev in deepcopy(devices).items():
        if sn in attributes:
            matches.append(
                Matches(
                    hostname=dev["hostname"],
                    sn=sn,
                    slug=dev["slug"],
                    old_site_name=attributes[sn]["value"],
                    rule_type="attributes",
                    rule_number=-1,
                )
            )
            devices.pop(sn)
    return matches


def _create_device_match(rule, device, idx):
    transformation = rule["transformation"] if rule["transformation"] != "none" else None
    return Matches(
        hostname=device["hostname"],
        slug=device["slug"],
        cloud_resource_id=device["resourceId"],
        old_site_name=device["sn"],
        sn=device["sn"],
        new_site_name=rule["siteName"],
        rule_type=rule["type"],
        rule_number=idx,
        rule_note=rule["note"],
        regex=rule.get("regex", None),
        transformation=transformation,
    )


def _get_cloud_inventory(ipf) -> dict:
    cloud = dict()
    for k, v in {"aws": "arn", "azure": "sn", "gcp": "resourceName"}.items():  # TODO: NIM-17070
        columns = {"sn"}
        columns.add(v)
        cloud.update(
            {_["sn"]: _[v] for _ in ipf.fetch_all(f"tables/cloud/vendors/{k}/inventory", columns=list(columns))}
        )
    return cloud


def map_devices_to_rules(ipf, snapshot_id: str = "$last"):
    ss = SiteSeparation(client=ipf)
    devices = {
        d["sn"]: d
        for d in ipf.inventory.devices.all(
            columns=["hostname", "sn", "siteName", "slug", "vendor"], snapshot_id=snapshot_id
        )
    }
    cloud = _get_cloud_inventory(ipf)
    {devices[_].update({"resourceId": cloud.get(_, None)}) for _ in devices}
    rules = ss.get_separation_rules()

    matches = check_attributes(ipf, devices) if rules["manualEnabled"] else list()

    for idx, rule in enumerate(rules["rules"]):
        if rule["type"] != "slug":
            data = ss.get_rule_matches(rule)
            for match in data["matched"]:
                if match["sn"] in devices:
                    matches.append(_create_device_match(rule, devices[match["sn"]], idx))
                    devices.pop(match["sn"], None)
        else:
            for device in devices.copy().values():
                if device["vendor"] in rule["enabledVendors"]:
                    # TODO: Include also cloud devices discovered via SSH or Vendor API
                    rule["siteName"] = device["siteName"]
                    matches.append(_create_device_match(rule, device, idx))
                    devices.pop(device["sn"], None)

    for sn, dev in devices.items():
        matches.append(
            Matches(
                hostname=dev["hostname"],
                sn=sn,
                slug=dev["slug"],
                cloud_resource_id=dev["resourceId"],
                old_site_name=dev["siteName"],
                rule_type="noMatchingRule",
            )
        )
    return [vars(m) for m in matches]



from copy import deepcopy
from typing import Optional, Union

from deepdiff import DeepDiff
from pydantic import BaseModel, field_validator, Field

from ipfabric.diagrams.output_models import Position
from .constants import (
    DEFAULT_NETWORK,
    DEFAULT_PATHLOOKUP,
    HIDDEN_DEV_TYPES,
)
from .graph_parameters import (
    Unicast,
    Multicast,
    Host2GW,
    Network,
)
from .graph_settings import NetworkSettings, PathLookupSettings, PathLookup, Overlay

REMOVE_PATH_PARAMS = ["pathLookupType", "networkMode", "pathLookupType", "groupBy", "type"]
PARAMS_IMPORTS = [
    "EntryPoint",
    "Layout",
    "Technologies",
    "Instance",
    "STPInstances",
    "OtherOptions",
    "Algorithm",
    "Overlay",
    "EdgeSettings",
    "GroupSettings",
    "Style",
    "PathLookup",
]


class SharedView(BaseModel):
    snapshot_id: str
    client_snapshot_id: str
    positions: dict
    settings: list[dict]
    params: dict
    hidden_devs: Optional[list] = Field(default_factory=list)
    hidden_nodes: Optional[list] = Field(default_factory=list)  # Not available in API
    collapsed_node_groups: Optional[list] = Field(default_factory=list)  # Not available in API
    path_lookup: Optional[dict] = Field(default_factory=dict)
    overlay: Optional[dict] = Field(default_factory=dict)

    @field_validator("settings")
    @classmethod
    def _settings(cls, v):
        for e in v:
            e.pop("id", None)
            for c in e.get("children", []):
                c.pop("id", None)
        return v

    @property
    def graph_type(self):
        if self.params["type"] == "topology":
            return self.params["type"]
        else:
            return self.params["pathLookupType"]

    def graph_model(self) -> Union[Unicast, Host2GW, Multicast, Network]:
        if self.graph_type == "topology":
            return Network(sites=self.params["paths"], technologies=self.params.get("technologies", None))
        params = deepcopy(self.params)
        [params.pop(_, None) for _ in REMOVE_PATH_PARAMS]
        params.update(params.pop("l4Options", {}))
        if self.graph_type == "unicast":
            return Unicast(**params)
        elif self.graph_type == "hostToDefaultGW":
            return Host2GW(**params)
        elif self.graph_type == "multicast":
            return Multicast(**params)

    def graph_settings(self):
        if self.graph_type == "topology":
            if DeepDiff(self.settings, DEFAULT_NETWORK) or self.hidden_devs != HIDDEN_DEV_TYPES:
                return NetworkSettings(edges=self.settings, hiddenDeviceTypes=self.hidden_devs)
            else:
                return None
        else:
            if DeepDiff(self.settings, DEFAULT_PATHLOOKUP) or self.path_lookup != vars(PathLookup()):
                return PathLookupSettings(edges=self.settings, pathLookup=self.path_lookup)
            else:
                return None

    @property
    def overlay_settings(self):
        if self.overlay.get("type", None) == "intent":
            return Overlay(intentRuleId=self.overlay["intentRule"]["id"])
        elif self.overlay.get("type", None) == "compare":
            return Overlay(snapshotToCompare=self.overlay["snapshotToCompare"])
        return None

    @property
    def positions_settings(self) -> dict[str, Position]:
        return {k: Position(**v) for k, v in self.positions.items()}

    def create_code(self, positions: bool = False):
        settings = self.graph_settings()
        params = self.graph_model()
        code = [f"parameters = {params.__repr_name__()}({params.__repr_str__(', ')})"]
        imports, e_imports = {params.__repr_name__()}, {_ for _ in PARAMS_IMPORTS if _ in str(params)}

        imports.update({"ICMP"} if getattr(params, "protocol", None) == "icmp" else {})
        e_imports.update({"Overlay"} if self.overlay_settings else {})

        tmp = "diagram = ipf.diagram.model(parameters=parameters"
        if settings:
            imports.add(settings.__repr_name__())
            code.append(f"settings = {settings.__repr_name__()}({settings.__repr_str__(', ')})")
            e_imports.update({_ for _ in PARAMS_IMPORTS if _ in str(settings)})
            tmp += ", graph_settings=settings"
        if e_imports:
            code.insert(0, f"from ipfabric.diagrams.input_models import {', '.join(e_imports)}")

        if positions:
            imports.add("Position")
            code.append(f"positions = {self.positions_settings}")
            tmp += ", positions=positions"
        code.insert(0, f"from ipfabric.diagrams import {', '.join(imports)}")

        if self.snapshot_id != self.client_snapshot_id:
            tmp += f", snapshot_id='{self.snapshot_id}'"
        if self.overlay_settings:
            tmp += f", overlay=Overlay({self.overlay_settings.__repr_str__(', ')})"
        tmp += ")"
        code.append(tmp)

        return code



from __future__ import annotations

import os
from typing import Optional, Union

from pydantic import BaseModel

PYDANTIC_EXTRAS = os.getenv("IPFABRIC_PYDANTIC_EXTRAS", "allow")


class BaseSecurity(BaseModel, extra=PYDANTIC_EXTRAS):
    name: Optional[str] = None
    type: Optional[str] = None
    original: Optional[str] = None


class DefaultAction(BaseSecurity, BaseModel, extra=PYDANTIC_EXTRAS):
    chainName: Optional[str] = None


class Reference(BaseSecurity, BaseModel, extra=PYDANTIC_EXTRAS):
    dictionary: str
    key: str


class ReferenceValue(BaseModel, extra=PYDANTIC_EXTRAS):
    reference: Reference


class ValueItem(BaseSecurity, BaseModel, extra=PYDANTIC_EXTRAS):
    min: Union[int, str]
    max: Union[int, str]
    version: Optional[int] = None


class ValueOriginalType(BaseSecurity, BaseModel, extra=PYDANTIC_EXTRAS):
    category: Optional[str] = None
    value: list[Union[ValueItem, ValueOriginalType, ReferenceValue]]


class ValueItemIpPort(BaseSecurity, BaseModel, extra=PYDANTIC_EXTRAS):
    toIp: ValueOriginalType
    toPort: Optional[ValueOriginalType] = None


class HeaderField(BaseModel, extra=PYDANTIC_EXTRAS):
    header: Optional[str] = None
    field: Optional[str] = None


class NexthopIp(BaseModel):
    ip: str
    ipLong: int


class NexthopInt(BaseModel):
    int: str


class PathLookupOption(BaseSecurity, HeaderField, BaseModel, extra=PYDANTIC_EXTRAS):
    matchedValue: Optional[Union[ValueOriginalType, ReferenceValue]] = None
    isDynamic: Optional[bool] = None
    value: Optional[Union[str, ValueOriginalType, list[ValueItemIpPort], ReferenceValue]] = None
    nexthopList: Optional[list[Union[NexthopIp, NexthopInt]]] = None


class LogOptions(BaseModel, extra=PYDANTIC_EXTRAS):
    logStart: bool
    logEnd: bool


class DiscoveryOptions(BaseModel, extra=PYDANTIC_EXTRAS):
    logOptions: LogOptions


class Action(BaseSecurity, BaseModel, extra=PYDANTIC_EXTRAS):
    pathLookupOptions: Optional[Union[list[PathLookupOption], list[ValueItemIpPort], str]] = None
    chainName: Optional[str] = None
    discoveryOptions: Optional[DiscoveryOptions] = None


class MaskResult(BaseModel, extra=PYDANTIC_EXTRAS):
    mask: int
    result: int


class Right(BaseSecurity, BaseModel, extra=PYDANTIC_EXTRAS):
    value: Optional[Union[str, int, MaskResult, list[Union[str, int, ValueItem]]]] = None
    reference: Optional[Reference] = None


class LeftRight(BaseSecurity, BaseModel, extra=PYDANTIC_EXTRAS):
    left: Optional[HeaderField] = None
    operator: Optional[str] = None
    right: Optional[Right] = None
    category: Optional[str] = None
    reference: Optional[Reference] = None
    description: Optional[str] = None


class Group(LeftRight, BaseModel, extra=PYDANTIC_EXTRAS):
    group: str
    name: Optional[str] = None
    items: Optional[list[Union[LeftRight, Group]]] = None
    item: Optional[Union[LeftRight, Group]] = None


class Chain(BaseModel, extra=PYDANTIC_EXTRAS):
    chainName: str
    chainType: str
    checkType: str


class Test(LeftRight, BaseModel, extra=PYDANTIC_EXTRAS):
    group: Optional[str] = None
    items: Optional[list[Union[LeftRight, Group, Chain]]] = None


class Category(BaseModel, extra=PYDANTIC_EXTRAS):
    category: str
    operator: str
    original: str


class Rule(BaseSecurity, BaseModel, extra=PYDANTIC_EXTRAS):
    action: Action
    active: bool
    natEvaluationType: Optional[str] = None
    test: Union[Test, Chain, Category]
    uid: Optional[str] = None
    description: Optional[str] = None
    sequenceNumber: Optional[int] = None


class RuleChain(BaseSecurity, BaseModel, extra=PYDANTIC_EXTRAS):
    defaultAction: DefaultAction
    rules: list[Rule]
    hideInUi: Optional[bool] = False


class SetField(BaseSecurity, BaseModel, extra=PYDANTIC_EXTRAS):
    category: Optional[str] = None
    description: Optional[str] = None
    value: list[ValueItem]


class StringArray(BaseSecurity, BaseModel, extra=PYDANTIC_EXTRAS):
    value: list[str]


class String(BaseSecurity, BaseModel, extra=PYDANTIC_EXTRAS):
    value: str


class NamedValues(BaseModel, extra=PYDANTIC_EXTRAS):
    numberSet: Optional[dict[str, SetField]] = None
    numberSetGroup: Optional[dict[str, Union[SetField, ValueOriginalType]]] = None
    stringSet: Optional[dict[str, SetField]] = None
    stringArray: Optional[dict[str, StringArray]] = None
    string: Optional[dict[str, String]] = None


class Machine(BaseModel, extra=PYDANTIC_EXTRAS):
    entryChainName: str
    ruleChains: dict[str, RuleChain]
    namedTests: Optional[dict[str, Union[LeftRight, Group]]] = None
    namedValues: Optional[NamedValues] = None


class Security(BaseModel, extra=PYDANTIC_EXTRAS):
    machineAcl: Optional[Machine] = None
    machineNat44: Optional[Machine] = None
    machinePbr: Optional[Machine] = None
    machineZones: Optional[Machine] = None
    zones: Optional[dict[str, list[str]]] = None


class SecurityModel(BaseModel, extra=PYDANTIC_EXTRAS):
    hostname: str
    security: Security



import logging
from copy import deepcopy
from typing import Any, Literal, Optional

from pydantic import Field, BaseModel

from ipfabric.tools.shared import raise_for_status

logger = logging.getLogger("ipfabric")


class SiteSeparation(BaseModel):
    client: Any = Field(None, exclude=True)

    def get_separation_rules(self):
        return self.client.get("settings/site-separation").json()

    def _post_rule(self, data):
        return raise_for_status(self.client.post("settings/site-separation/test-regex", json=data)).json()

    def get_rule_matches(self, rule):
        rule = deepcopy(rule)
        [rule.pop(key, None) for key in ["id", "note"]]
        return self._post_rule(rule)

    def _get_matches(
        self, regex: str, transformation, type: str, site_name: Optional[str] = None, cloud: Optional[bool] = None
    ):
        transformation = transformation.lower()
        if transformation not in ["uppercase", "lowercase", "none"]:
            raise SyntaxError('Transformation type is not in ["uppercase", "lowercase", "none"].')
        rule = {"regex": regex, "transformation": transformation, "siteName": site_name, "type": type}
        if isinstance(cloud, bool):
            rule["applyToCloudInstances"] = cloud
        return self._post_rule(rule)

    def get_hostname_matches(
        self, regex: str, transformation: Literal["uppercase", "lowercase", "none"], site_name: Optional[str] = None
    ):
        return self._get_matches(regex, transformation, "regexHostname", site_name)

    def get_snmp_matches(
        self, regex: str, transformation: Literal["uppercase", "lowercase", "none"], site_name: Optional[str] = None
    ):
        return self._get_matches(regex, transformation, "regexSnmpLocation", site_name)

    def get_cloud_id_matches(
        self,
        regex: str,
        transformation: Literal["uppercase", "lowercase", "none"],
        site_name: Optional[str] = None,
        apply_to_cloud: bool = False,
    ):
        return self._get_matches(regex, transformation, "regexCloudResourceId", site_name, apply_to_cloud)
